"""
Created on Wed Apr 28 23:03:36 2021

@author: Haotian Teng
"""

import os
import sys
import h5py
import toml
import numpy as np
import argparse
from matplotlib import pyplot as plt
from itertools import compress
import seaborn as sns
from tqdm import tqdm
from xron.utils.seq_op import fast5_iter,norm_by_noisiest_section,diff_norm_by_noisiest_section,diff_norm_fixing_deviation
from xron.utils.align import MetricAligner
from Bio.Seq import Seq
from functools import partial
from boostnano.boostnano_model import CSM
from boostnano.boostnano_eval import evaluator
import inspect
import os
from pathlib import Path
from xron.nrhmm.prepare_data import Extractor

alt_map = {'ins':'0','M':'A','U':'T'}
complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'} 
MIN_READ_SEQ_LEN = 100 #The filter of minimum sequence length.
RNA_FILTER_CONFIG = {"min_rate":25,
                     "min_seq_len":3, #this is in term of each chunk
                     "max_gap_allow":2000,
                     "min_quality_score":0.85, #the minimum quality score for a chunk to be included into traning set.
                     "max_mono_prop":0.6}

DNA_FILTER_CONFIG = {"min_rate":2,
                     "min_seq_len":7,
                     "max_gap_allow":400,
                     "min_quality_score":0.8, #the minimum quality score for a chunk to be included into traning set.
                     "max_mono_prop":0.8}
def reverse_complement(seq):    
    return str(Seq(seq).reverse_complement())

def chop(arr,chunk_length,padding = True,pad_values = 0):
    read_len = len(arr)
    chunks = np.split(arr,np.arange(0,read_len,chunk_length))[1:]
    length = [len(x) for x in chunks]
    if padding:
        last_chunk = chunks[-1]
        chunks[-1]= np.pad(last_chunk,(0,chunk_length-len(last_chunk)),'constant',constant_values = (pad_values,pad_values))
    return chunks,length

def clean_repr(seq):
    for k,v in alt_map.items():
        seq = seq.replace(k,v)
    return seq

def retrive_seq(seq_h,event_stride):
    moves = np.asarray(seq_h['BaseCalled_template']['Move'])
    try:
        seq = np.asarray(seq_h['BaseCalled_template']['Fastq']).tobytes().decode('utf-8').split('\n')[1]
    except:
        seq = str(np.asarray(seq_h['BaseCalled_template']['Fastq']).astype(str)).split('\n')[1]
    pos = np.repeat(np.cumsum(moves)-1,repeats = event_stride).astype(np.int32)
    return seq,pos
    
def filt(filt_config,chunks,seq,seq_len,durations,meta,*args):
    n = chunks.shape[0]
    segment_len = chunks.shape[1]
    print("Origin %d chunks in total."%(chunks.shape[0]))
    
    ### Filter out by length
    max_seq_len = np.int(segment_len/filt_config['min_rate'])
    mask_min = seq_len>filt_config["min_seq_len"]
    mask_max = seq_len<max_seq_len
    ### Filter by quality score
    qs_mask = np.asarray([x!='X' for x in seq]) # 'X' is the mark of low quality chunks.
    qs_penc = 100*sum(np.logical_not(qs_mask))/n
    print("%.2f"%(qs_penc),"% chunks has been filtered out because of quality score.")

    ### Filter by boostnano
    bn_mask = np.asarray([x!='Y' for x in seq]) # 'Y' means current chunks is from a polyA tail or adapter sequence.
    bn_penc = 100*sum(np.logical_not(bn_mask))/n
    print("%.2f"%(bn_penc),"% chunks has been filtered out because they are polyA tail or adapter.")    
    
    ### Filter by gap
    gap_mask = np.asarray([x!='P' for x in seq]) # 'P' is the mark of chunks with large gap.
    gap_perc = 100*sum(np.logical_not(gap_mask))/n
    print("%.2f"%(gap_perc),"% chunks has been filtered out because of large gap.")

    print("%.2f"%(100*sum(np.logical_not(mask_min))/n - gap_perc - qs_penc - bn_penc),"% chunks are filted out by minimum sequence length filter.")
    print("%.2f"%(100*sum(np.logical_not(mask_max))/n),"% chunks are filted out by maximum sequence length filter.")
    mask = (mask_min*mask_max*qs_mask*gap_mask*bn_mask).astype(bool)
    
    
    ### Filter out by monopolization
    mono_mask = np.asarray([max(np.unique(list(x),return_counts = True)[1])/y<filt_config['max_mono_prop'] for x,y in zip(seq[mask],seq_len[mask])])
    print("%.2f"%(100*sum(np.logical_not(mono_mask))/sum(mask)),"% chunks are filted out by mono filter.")
    ### Filter by unrecognized nucleotide type
    type_mask = np.asarray([len(x.replace('A','').replace('G','').replace('C','').replace('T',''))==0 for x in seq[mask]])
    print("%.2f"%(100*sum(np.logical_not(type_mask))/sum(mask)),"% chunks are filted out by nucleotide filter.")
    
    mono_mask = (mono_mask*type_mask).astype(bool) 
    
    pass_chunks = chunks[mask][mono_mask]
    print("In total %.2f"%(100*len(pass_chunks)/len(chunks)),"% of reads pass the filters.")
    filtered_args = [arg[mask][mono_mask] for arg in args]
    if filtered_args:
        return pass_chunks,seq[mask][mono_mask],seq_len[mask][mono_mask],durations[mask][mono_mask],list(compress(compress(meta,mask),mono_mask)),filtered_args
    else:
        return pass_chunks,seq[mask][mono_mask],seq_len[mask][mono_mask],durations[mask][mono_mask],list(compress(compress(meta,mask),mono_mask))
    
def rna_filt(chunks,seq,seq_len,durations,meta,*args):
    return partial(filt,RNA_FILTER_CONFIG)(chunks,seq,seq_len,durations,meta,*args)

def dna_filt(chunks,seq,seq_len,durations,meta,*args):
    return partial(filt,DNA_FILTER_CONFIG)(chunks,seq,seq_len,durations,meta,*args)

def extract(args):
    if args.mode == 'dna':
        FILTER_CONFIG = DNA_FILTER_CONFIG
    else:
        FILTER_CONFIG = RNA_FILTER_CONFIG
    if args.write_correction:
        m = 'r+'
    else:
        m = 'r'
    if args.extract_kmer:
        extractor = Extractor(k=args.kmer_size,alphabeta = args.alphabeta)
    iterator = fast5_iter(args.input_fast5,mode = m)
    if args.diff_sig:
        if args.config['fixed_deviation']:
            norm_func = diff_norm_fixing_deviation
        else:
            norm_func = diff_norm_by_noisiest_section
    else:
        norm_func = norm_by_noisiest_section
    if args.extract_seq:
        print("Read reference genome.")
        aligner = MetricAligner(args.reference,options = '-x ont2d')
    print("Loading BoostNano model.")
    project_f = os.path.dirname(os.path.dirname(inspect.getfile(CSM)))
    model_f = os.path.join(project_f,'BoostNano','model')
    net = CSM()
    boostnano_evaluator = evaluator(net,model_f)
    print("Begin processing the reads.")
    meta_info,chunks,seqs,meds,mads,meths,offsets,scales,read_ids,kmers,durations = [],[],[],[],[],[],[],[],[],[],[]
    if args.mode == "rna" or args.mode == "rna_meth":
        reverse_sig = True
    fail_read_count = {"No basecall":0,
                       "Alignment failed":0,
                       "Read too short":0,
                       "Sequence length is inconsistent with signal length":0,
                       "Kmer extraction failed":0,
                       "Processed":0,}
    
    ### Debug code
    qss = []
    ###
    loop_obj = tqdm(iterator)
    for read_h,signal,fast5_f,read_id in loop_obj:
        postfix_str = "no entry: %d, alignment failed: %d, read too short: %d, inconsistent length: %d, kmer fail:%d, processed:%d"%(fail_read_count['No basecall'],
                                                                                                                                  fail_read_count['Alignment failed'],
                                                                                                                                  fail_read_count['Read too short'],
                                                                                                                                  fail_read_count['Sequence length is inconsistent with signal length'],
                                                                                                                                  fail_read_count['Kmer extraction failed'],
                                                                                                                                  fail_read_count['Processed'])
        loop_obj.set_postfix_str(postfix_str)
        read_len = len(signal)
        original_signal = signal.astype(np.float32)
        signal,med,mad = norm_func(signal)
        signal = signal.astype(np.float16)
        if reverse_sig:
            signal = signal[::-1]
        if args.extract_seq: 
            try:
                seq, pos = retrive_seq(read_h['Analyses/Basecall_1D_%s'%(args.basecall_entry)],
                                   args.stride)
                basecall_entry = args.basecall_entry
            except KeyError:
                if args.alternative_entry is not None:
                    try:
                        seq, pos = retrive_seq(read_h['Analyses/Basecall_1D_%s'%(args.alternative_entry)],
                                           args.stride)
                        basecall_entry = args.alternative_entry
                    except KeyError:
                        print("No basecall information was found in entry Basecall_1D_%s of read %s of %s, skip the read."%(args.basecall_entry,read_id,fast5_f))
                        fail_read_count["No basecall"]+=1
                        continue
                else:
                    print("No basecall information was found in entry Basecall_1D_%s of read %s of %s and no alternative entry is set, skip the read."%(args.basecall_entry,read_id,fast5_f))
                    fail_read_count["No basecall"]+=1
                    continue
            assert read_h['Analyses/Basecall_1D_%s/Summary/basecall_1d_template'%(basecall_entry)].attrs['block_stride'] == args.stride
            if args.write_correction:
                if not "Segmentation_%s"%(basecall_entry) in read_h['Analyses']:
                    read_h['Analyses/'].create_group("Segmentation_%s"%(basecall_entry))
                if not "Reference_corrected" in read_h['Analyses/Segmentation_%s/'%(basecall_entry)]:
                    read_h['Analyses/Segmentation_%s/'%(basecall_entry)].create_group("Reference_corrected")
                else:
                    del read_h['Analyses/Segmentation_%s/Reference_corrected'%(basecall_entry)]
                    read_h['Analyses/Segmentation_%s/'%(basecall_entry)].create_group("Reference_corrected")
            if len(seq) < MIN_READ_SEQ_LEN:
                fail_read_count["Read too short"]+=1
                continue
            if seq.count('A')+seq.count('M') == 0:
                meths.append(np.nan)
            else:
                meths.append(seq.count('M')/(seq.count('A')+seq.count('M')+1e-8))
            seq = clean_repr(seq) #M->A, U->T
            hits,ref_seq,ref_idx,qs = aligner.ref_seq(seq,exclude_negative=True) #ref_idx: the ith ref_idx R_i means the ith base in reference sequence is the R_i base in basecalled sequence
            if not hits:
                fail_read_count['Alignment failed'] += 1
                continue
            if args.mode == 'rna' or args.mode == 'rna_meth':
                (decoded,path,locs) = boostnano_evaluator.eval_sig(original_signal,1000)
                locs = read_len - locs
            assert np.all(np.diff(ref_idx)>=0)
            try:
                start = int(read_h['Analyses/Segmentation_%s/Summary/segmentation'%(basecall_entry)].attrs['first_sample_template'])
            except:
                start = 0
            if reverse_sig and args.rev_move and start>0:
                signal = signal[:-start]
            if args.rev_move and reverse_sig:
                pos = pos[::-1]
                pos = pos[0] - pos 
            else:
                signal = signal[start:]
            if abs(len(signal)-len(pos)) > (min(len(signal),len(pos))/args.stride):
                print(fast5_f,read_id,len(pos),len(signal))
                print("The signal length is %d and position length is (%d) for read %s of %s, check if the stride is correct."%(len(signal),len(pos),read_id,fast5_f))
                fail_read_count["Sequence length is inconsistent with signal length"] +=1
                continue
            if len(signal)>len(pos):
                if args.padding:
                    #Insert values on random selected indexs.
                    padding = len(signal) - len(pos)
                    idxs = np.random.choice(len(pos),padding,replace = False)
                    pos = np.insert(pos,idxs,pos[idxs])
                else:
                    signal = signal[:len(pos)]
            else:
                pos = pos[:len(signal)]
            if len(signal) == 0:
                continue
            read_len = len(pos)
            if args.write_correction:
                ref_idx_aligned = ref_idx[ref_idx<=pos[-1]]
                ref_seq_aligned = ref_seq[:len(ref_idx_aligned)]
                if 'N' in ref_seq_aligned:
                    fail_read_count["Kmer extraction failed"] += 1
                    continue
                qs_aligned = qs[:len(ref_idx_aligned)]
                ref_sig_idx = [np.where(pos == x)[0][0] for x in ref_idx_aligned] #The ith ref_sig_idx REF_i means the ith base in reference sequence is the REF_i signal point at the reversed signal.
                read_h['Analyses/Segmentation_%s/Reference_corrected'%(basecall_entry)].create_dataset("ref_sig_idx",data = ref_sig_idx)
                read_h['Analyses/Segmentation_%s/Reference_corrected'%(basecall_entry)].create_dataset("ref_seq",data = ref_seq_aligned)
                read_h['Analyses/Segmentation_%s/Reference_corrected'%(basecall_entry)].create_dataset("map_score",data = qs_aligned)
                if args.extract_kmer:
                    try:
                        _,kmer_seqs = extractor.kmer_decode(ref_seq_aligned, signal, np.asarray(ref_sig_idx), padded = True)
                    except ValueError:
                        fail_read_count["Kmer extraction failed"]+=1
                        continue
            fail_read_count["Processed"] += 1
            for x in np.arange(0,read_len,args.chunk_len):
                if args.mode == "rna" or args.mode == "rna_meth":
                    if x+args.chunk_len > locs[0]:
                        seqs.append('Y')
                        continue
                s,e = pos[x:x+args.chunk_len][[0,-1]]
                mask = (ref_idx>=s)&(ref_idx<=e)
                qs_mask = qs[mask]
                if sum(mask) > 0:
                    qss.append(sum(qs_mask)/len(qs_mask))
                    if sum(qs_mask)/len(qs_mask) <= FILTER_CONFIG["min_quality_score"]:
                        seqs.append('X')
                        continue
                    if np.max(np.unique(pos[x:x+args.chunk_len],return_counts = True)[1])>FILTER_CONFIG["max_gap_allow"]:
                        seqs.append('P')
                        continue
                    r_s,r_e = np.where(mask)[0][[0,-1]]
                    # print("Basecall sequence:%s"%(basecall_seq[s:e+1]))
                    seq = ref_seq[r_s:r_e+1]
                    # print("Aligned seqeuence:%s"%(seq))
                    if args.mode == 'rna-meth':
                        seq.replace('A','M')
                    elif args.mode == 'rna':
                        seq.replace('M','A')
                    seqs.append(seq)
                else:
                    seqs.append('')
        meds.append(med)
        mads.append(mad)
        if 'channel_id' in read_h:
            offsets.append(read_h['channel_id'].attrs['offset'])
            scales.append(read_h['channel_id'].attrs['range'])
        else:
            offsets.append(read_h['UniqueGlobalKey/channel_id'].attrs['offset'])
            scales.append(read_h['UniqueGlobalKey/channel_id'].attrs['range'])
        read_ids.append(read_id)
        curr_chunks,curr_duration = chop(signal,args.chunk_len,pad_values = 0)
        if args.extract_kmer:
            curr_kmers,_ = chop(kmer_seqs,args.chunk_len,pad_values = -1)
            kmers += curr_kmers
        chunks += curr_chunks
        durations += curr_duration
        meta_info += [(fast5_f,read_id,str(args.chunk_len),str(args.stride))]*len(curr_chunks)
        if args.max_n and (args.max_n > 0) and (len(chunks)>args.max_n):
            chunks = chunks[:args.max_n]
            seqs = seqs[:args.max_n]
            meta_info = meta_info[:args.max_n]
            kmers = kmers[:args.max_n]
            durations = durations[:args.max_n]
            break
    for key,val in fail_read_count.items():
        print(key,':',val)
    if len(chunks) == 0:
        raise ValueError("No chunk is added to dataset, check the setting.")
    chunks = np.stack(chunks,axis = 0)
    durations = np.asarray(durations,dtype = np.int)
    if args.extract_kmer:
        kmers = np.stack(kmers,axis = 0)
    print("Average median value %f"%(np.mean(meds)))
    print("Average median absolute deviation %f"%(np.mean(mads)))
    print("Average methylation proportion %f"%(np.nanmean(meths)))
    if args.extract_seq:
        seq_lens = [len(i) for i in seqs]
        seqs = np.array(seqs)
        seq_lens = np.array(seq_lens)
        filt_func = dna_filt if args.mode == "dna" else rna_filt
        if args.extract_kmer:
            chunks,seqs,seq_lens,durations,meta_info,more = filt_func(chunks,seqs,seq_lens,durations,meta_info,kmers)
            kmers = more[0]
        else:
            chunks,seqs,seq_lens,durations,meta_info = filt_func(chunks,seqs,seq_lens,durations,meta_info)
        np.save(os.path.join(args.output,'seqs.npy'),seqs)
        np.save(os.path.join(args.output,'seq_lens.npy'),seq_lens)
    if args.extract_kmer:
        np.save(os.path.join(args.output,'kmers.npy'),kmers)
    np.save(os.path.join(args.output,'mm.npy'),(mads,meds,meths,offsets,scales,read_ids))
    np.savetxt(os.path.join(args.output,'meta.csv'),meta_info,fmt="%s")
    np.save(os.path.join(args.output,'chunks.npy'),chunks)
    np.save(os.path.join(args.output,'durations.npy'),durations)
    config_file = os.path.join(args.output,'config.toml')
    config_modules = [x for x in args.__dir__() if not x .startswith('_')][::-1]
    config_dict = {x:getattr(args,x) for x in config_modules}
    config_dict['FILTER_CONFIG'] = FILTER_CONFIG
    config_dict['minimum_sequence_length']:MIN_READ_SEQ_LEN
    config_dict["chunk_len"] = args.chunk_len
    if args.extract_kmer:
        config_dict["k"] = args.kmer_size
        config_dict["alphabeta"] = args.alphabeta
        config_dict['kmer2idx_dict'] = extractor.kmer2idx_dict
        config_dict['idx2kmer'] = extractor.idx2kmer
    plt.figure()
    sns.distplot(qss)
    plt.xlabel("Quality score")
    plt.ylabel("Distribution")
    plt.savefig(os.path.join(args.output,"qs_dist.png"))
    with open(config_file,'w+') as f:
        toml.dump(config_dict,f)

def add_arguments(parser):
    parser.add_argument('-i', 
                        '--input_fast5', 
                        required = True,
                        help="File path or Folder path to the fast5 file.")
    parser.add_argument('-o',
                        '--output',
                        required = True,
                        help="Output folder.")
    parser.add_argument('--chunk_len',
                        default = 4000,
                        type=int,
                        help="The lenght of the segment in chunk.")
    parser.add_argument('--max_n',
                        default = None,
                        type=int,
                        help="The maximum number of the segments to be extracted")
    parser.add_argument("--extract_seq",
                        action = "store_true",  
                        help = "If the sequence information is going to be\
                            extracted.")
    parser.add_argument("--extract_kmer",
                        action = "store_true",
                        help = "If the kmer information is going to be extracted.")
    parser.add_argument("--kmer_size","-k",
                        default = None,
                        type = int,
                        help = "The length of the kmer to be extracted.")
    parser.add_argument("--write_correction",
                        action = "store_true",
                        dest = "write_correction",
                        help = "If write correcting alignment to the original fast5 file.")
    parser.add_argument('--basecall_entry',
                        default = "000",
                        help="The entry number in /Analysis/ to look into, for\
                            example 000 means looking for Basecall_1D_000.")
    parser.add_argument('--alternative_entry',
                        default = None,
                        type = str,
                        help="If basecall information is not found in the basecall entry, look into this alternative entry.")
    parser.add_argument('--stride',
                        default = 12,
                        type = int,
                        help = "The length of stride used in basecall model,\
                        for guppy RNA fast, this number is 12, for guppy RNA\
                        hac model, this number is 10, for xron this number is\
                        5 or 11."
                        )
    parser.add_argument('--padding',
                        type = bool,
                        default = True,
                        help = "Padding the position to make the position has same length, otherwise cut the signal.")
    parser.add_argument('--move_direction',
                        default = 'forward',
                        help = "The direction of the output Move matrix, for Guppy, it's forward, for Xron it's backward.")
    parser.add_argument('--basecaller',
                        default = None,
                        help = "The basecaller setting, default is None, can be xron, guppy and guppy_fast, set this argument will override args.stride and args.move_direction.")
    parser.add_argument('--reference',
                        default = None,
                        help = "The reference genome, it's required when\
                        extract_seq is set to True")
    parser.add_argument('--mode',
                        default = "rna",
                        help = "Can be one of this mode: rna, dna, rna-meth")
    parser.add_argument('--fix_d',action="store_true",
                        dest = "fix_d",
                        help = "Use a fix deviation to normalize the signal.")

def post_args(FLAGS):
    XRON_CONFIG = {"stride":11,
                   "padding":True, #If we padding the position to the same length as signal.
                   "differential_signal":False,
                   "forward_move_matrix":False,#If the move matrix is count on backward signal or forward signal, True means the move matrix is along the forward signal, so the move matrix should be reveresed first as the signal is reveresed before processing.
                   "fixed_deviation":FLAGS.fix_d} 
    GUPPY_CONFIG = {"stride":10,
                    "chunk_len":2000, #The length of the chunk when basecall.
                    "padding":False,
                    "forward_move_matrix":True,
                    "differential_signal":False}
    GUPPY_FAST_CONFIG = {"stride":12,
                         "padding":False,
                         "chunk_len":2000, #The length of the chunk when basecall.
                         "forward_move_matrix":True,
                         "differential_signal":False}
    config_dict = {"xron":XRON_CONFIG,
                   "guppy":GUPPY_CONFIG,
                   "guppy_fast":GUPPY_FAST_CONFIG}
    if FLAGS.basecaller:
        config = config_dict[FLAGS.basecaller]
        FLAGS.stride = config["stride"]
        FLAGS.rev_move = config["forward_move_matrix"]
        FLAGS.diff_sig = config["differential_signal"]
        FLAGS.padding = config['padding']
    else:
        print("Basecaller is not defined, initialize the configuration with guppy fast model.")
        config = config_dict["guppy_fast"]
        FLAGS.rev_move = FLAGS.move_direction
    FLAGS.config = config
    FLAGS.alphabeta = "ACGTM" if FLAGS.mode == "rna_meth" else "ACGT"
    assert FLAGS.mode in ["rna","dna","rna_meth"], "Mode can only be one of rna, dna, rna_meth"
    if FLAGS.extract_kmer:
        assert FLAGS.kmer_size is not None, "Please specify the kmer length when extract kmer is enable."
        assert FLAGS.write_correction, "Please enable write_correction when extract kmer is enable."
    if FLAGS.write_correction:
        assert FLAGS.reference is not None, "Please specify the reference genome when write_correction is enable."
        assert FLAGS.extract_seq, "Please enable extract_seq when write_correction is enable."
    if FLAGS.extract_seq:
        if not FLAGS.reference:
            raise ValueError("Reference genome is required when extract the \
                             sequence.")
    os.makedirs(FLAGS.output,exist_ok = True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='xron',   
                                     description='A Unsupervised Nanopore basecaller.')
    add_arguments(parser)
    FLAGS = parser.parse_args(sys.argv[1:])
    post_args(FLAGS)
    extract(FLAGS)





"""
This script is used to resquiggle the reads that have different tandem repeat number than the reference.
"""
import os
import toml
import argparse
import numpy as np
from functools import partial
from tqdm import tqdm
from multiprocessing import Pool
from multiprocessing import Queue
from xron.nrhmm.kmer2seq import fixing_looping_path
import multiprocessing as mp
from time import sleep 
from multiprocessing import Process
from multiprocessing import Manager
from multiprocessing import Queue
from tqdm import tqdm
import queue #In Python 2.7, this would be Queue

DESIRED_TIMEOUT = 1
MAX_RESULT_QUEUE_SIZE = 1000
SLEEPING_TIME = 0.01
def worker(job_queue,args,result_queue):
    while True:
        try:
            i,seq,path,duration = job_queue.get(timeout = DESIRED_TIMEOUT)
        except queue.Empty:
            return
        #Do the job here
        if len(seq) < args.min_seq_len:
            result_queue.put({i:(None,None)})
        else:
            fixed_path, fixing = fixing_looping_path_func(path[:duration],seq)
            # while result_queue.qsize() > MAX_RESULT_QUEUE_SIZE:
            #     sleep(SLEEPING_TIME)
            result_queue.put({i:(fixed_path,fixing)})

def main(args):
    seqs,path,durations,config = load_data(args.input_folder)
    manager = Manager()
    filequeue = manager.Queue()
    result_queue = manager.Queue(maxsize = MAX_RESULT_QUEUE_SIZE)
    results_dict = {"Fixed":0, "No need to fix":0, "Fix failed":0}
    max_threads_number = mp.cpu_count()-1 if args.threads is None else args.threads #1 thread is used for the main process
    all_proc = []
    file_number = len(path)
    for i in range(file_number):
        filequeue.put((i,seqs[i],path[i],durations[i]))
    for i in range(max_threads_number):
        p = Process(target = worker, args = (filequeue,args,result_queue))
        all_proc.append(p)
        p.start()
    to_finished = list(np.arange(file_number))
    with tqdm() as t:   
     t.total = file_number
     t.set_description("Fixing the path of control data.")
     while len(to_finished):
        try:
            result = result_queue.get(timeout = DESIRED_TIMEOUT)
            i = list(result.keys())[0]
            fixed_path, fixing = result[i]
            if fixing is None:
                results_dict["Fix failed"] += 1
            elif fixing:
                results_dict['Fixed'] += 1
                path[i][:durations[i]] = fixed_path
            else:
                results_dict['No need to fix'] += 1
            elements_in_queue = result_queue.qsize()
            t.set_description(f"Fixed {results_dict['Fixed']}, No need to fix {results_dict['No need to fix']}, Skip because sequence is too short {results_dict['Fix failed']}, elements in queue {elements_in_queue}")
            if i in to_finished:
                to_finished.remove(i)
            else:
                raise ValueError(f"The index {i} is not in the to_finished list.")
            t.update()
        except queue.Empty:
            continue
        except Exception as e:
            print(e)
            break
    print("Saving the delooped result.")
    np.save(os.path.join(args.input_folder,"path_fix.npy"),path)
    print("Finished")
    for i,p in enumerate(all_proc):
        p.join()

def load_data(input_folder):
    seqs = np.load(os.path.join(input_folder,"seqs.npy"))
    path = np.load(os.path.join(input_folder,"path.npy"))
    duration = np.load(os.path.join(input_folder,"durations.npy"))
    config = toml.load(os.path.join(input_folder,"config.toml"))
    return seqs,path,duration,config

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i","--input_folder",help = "The folder that contains the data to be fixed.")
    parser.add_argument("--modified_base",default = "M",help = "The modified base.")
    parser.add_argument("--canonical_base",default = "A",help = "The canonical base.")
    parser.add_argument("--threads",default = None,type = int,help = "The number of threads to use, default is None which means using all the threads.")
    parser.add_argument("--min_seq_len",default = 7,type = int,help = "The minimum length of the sequence to be fixed.")
    args = parser.parse_args()
    config = toml.load(os.path.join(args.input_folder,"config.toml"))
    fixing_looping_path_func = partial(fixing_looping_path,idx2kmer = config['idx2kmer'],modified_base=args.modified_base,canonical_base=args.canonical_base)
    main(args)


"""
This script train a VQ-VAE style embedding network on Nanopore sequencing signal.
@author: Haotian Teng
"""
import os 
import sys
import torch
import argparse
import numpy as np
from typing import Union,List
from itertools import chain
from torchvision import transforms
import torch.utils.data as data
from torch.utils.data.dataloader import DataLoader
from xron.xron_input import Dataset, ToTensor, NumIndex
from xron.utils.prepare_chunk import rna_filt,dna_filt
from xron.xron_train_base import Trainer, DeviceDataLoader, load_config
from xron.xron_model import REVCNN,DECODER_CONFIG,CRNN,CRITIC_CONFIG,CRITIC,MM_CONFIG,MM
from xron.xron_label import MetricAligner
from torch.distributions.one_hot_categorical import OneHotCategorical as OHC

class VQVAETrainer(Trainer):
    def __init__(self,
                 train_dataloader:DataLoader,
                 encoder:CRNN,
                 decoder:REVCNN,
                 mm:MM,
                 config:Union[DECODER_CONFIG,MM_CONFIG],
                 device:str = None,
                 eval_dataloader:DataLoader = None):
        """

        Parameters
        ----------
        train_dataloader : DataLoader
            Training dataloader.
        encoder: CRNN
            A Convolutional-Recurrent Neural Network
        decoder : REVCNN
            REVCNN decoder
        mm: MM
            Markov Model instance.
        device: str
            The device used to train the model, can be 'cpu' or 'cuda'.
            Default is None, use cuda device if it's available.
        config: Union[DECODER_CONFIG,MM_CONFIG]
            A CONFIG class contains unsupervised training configurations. Need 
            to contain at least these parameters: keep_record, device and 
            grad_norm.
        eval_dataloader : DataLoader, optional
            Evaluation dataloader, if None training dataloader will be used.
            The default is None.

        """
        super().__init__(train_dataloader=train_dataloader,
                         nets = {"encoder":encoder,
                                 "decoder":decoder,
                                 "mm":mm},
                         config = config,
                         device = device,
                         eval_dataloader = eval_dataloader)
        self.train_config = config.TRAIN
        self.global_step = 0
        self.score_average = 0
        self.mse_loss = torch.nn.MSELoss(reduction = "mean")
        self.records = {'rc_losses':[],
                        'rc_valid':[],
                        'embedding_loss':[],
                        'commitment_loss':[]}
    @property
    def encoder(self):
        return self.nets["encoder"]
    
    @property
    def decoder(self):
        return self.nets["decoder"]

    @property
    def mm(self):
        return self.nets["mm"]
    
    def train(self,
              epoches:int,
              optimizers:List[torch.optim.Optimizer],
              save_cycle:int,
              save_folder:str):
        """
        Train the encoder-decodr nets.

        Parameters
        ----------
        epoches : int
            Number of epoches to train.
        optimizers : List[torch.optim.Optimizer]
            A list of three optimizers, the first one is optimizer training the
            encoder parameters, the second one for decoder parameters and the
            third one is the optimizer for the embedding.
        save_cycle : int
            Save every save_cycle batches.
        save_folder : str
            The folder to save the model and training records.

        Returns
        -------
        None.

        """
        self.save_folder = save_folder
        self._save_config()
        records = self.records
        for epoch_i in range(epoches):
            for i_batch, batch in enumerate(self.train_ds):
                losses = self.train_step(batch)
                loss = losses[0] + losses[1] + self.train_config['alpha']*losses[2]
                for opt in optimizers:
                    opt.zero_grad()
                loss.backward()
                for opt in optimizers:
                    opt.step()
                if (self.global_step+1)%save_cycle==0:
                    self.save()
                    eval_i,valid_batch = next(enumerate(self.eval_ds))
                    with torch.no_grad():
                        valid_rc = self.valid_step(valid_batch)
                    records["rc_valid"].append(valid_rc.detach().cpu().numpy()[()])
                    records['rc_losses'].append(losses[0].detach().cpu().numpy()[()])
                    records['embedding_loss'].append(losses[1].detach().cpu().numpy()[()])
                    records['commitment_loss'].append(losses[2].detach().cpu().numpy()[()])
                    print("Epoch %d Batch %d, rc_loss %f, embedding_loss %f, validation rc %f"%(epoch_i, i_batch, losses[0], losses[1],valid_rc))
                    self._update_records()
                losses = None
                torch.nn.utils.clip_grad_norm_(self.parameters, 
                                               max_norm=self.grad_norm)
                self.global_step +=1
        
    def train_step(self,batch):
        encoder = self.encoder
        decoder = self.decoder
        embedding = self.mm.level_embedding
        signal = batch['signal']
        e,q,e_shadow,sg_q,sg_e = encoder.forward_embedding(signal,embedding = embedding) 
        rc_signal = decoder.forward(e).permute([0,2,1]) #[N,L,C] -> [N,C,L]
        rc_loss = self.mse_loss(rc_signal,signal)
        embedding_loss = self.mse_loss(sg_q,e_shadow)
        commitment_loss = self.mse_loss(sg_e,q)
        return rc_loss, embedding_loss, commitment_loss
    
    def valid_step(self,batch):
        rc_loss,_,_ = self.train_step(batch)
        return rc_loss
    
def main(args):
    class CTC_CONFIG(MM_CONFIG):
        CTC = {"beam_size":5,
               "beam_cut_threshold":0.05,
               "alphabeta": "ACGTM",
               "mode":"rna"}
    class TRAIN_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "epsilon":0.1,
                 "epsilon_decay":0,
                 "alpha":1.0, #Entropy loss scale factor
                 "keep_record":5,
                 "decay":args.decay,
                 "diff_signal":args.diff}
        
    config = TRAIN_CONFIG()
    config.PORE_MODEL["N_BASE"] = len(config.CTC["alphabeta"])
    print("Read chunks and sequence.")
    chunks = np.load(args.chunks,allow_pickle = True,mmap_mode= 'r')
    print("Construct and load the model.")
    model_f = args.model_folder
    dataset = Dataset(chunks,seq = None,seq_len = None,transform = transforms.Compose([ToTensor()]))
    loader = data.DataLoader(dataset,batch_size = args.batch_size,shuffle = True, num_workers = 4)
    DEVICE = args.device
    loader = DeviceDataLoader(loader,device = DEVICE)
    if args.retrain:
        config_old = load_config(os.path.join(model_f,"config.toml"))
        config_old.TRAIN = config.TRAIN #Overwrite training config.
        config = config_old
    if args.config:
        config_old = load_config(args.config)
        config_old.TRAIN = config.TRAIN
        config = config_old
    encoder = CRNN(config)
    decoder = REVCNN(config)
    mm = MM(config)
    t = VQVAETrainer(loader,encoder,decoder,mm,config)
    if args.retrain:
        t.load(model_f)
    lr = args.lr
    epoches = args.epoches
    opt = torch.optim.Adam(chain(t.encoder.parameters(),t.decoder.parameters(),t.mm.level_embedding.parameters()),lr = lr)
    COUNT_CYCLE = args.report
    print("Begin training the model.")
    t.train(epoches,[opt],COUNT_CYCLE,model_f)
        
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training model with chunks and sequence file')
    parser.add_argument('-i', '--chunks', required = True,
                        help = "The .npy file contain chunks.")
    parser.add_argument('-o', '--model_folder', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument('--config', default = None,
                        help = "The configure file used to train.")
    parser.add_argument('--device', default = 'cuda',
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--lr', default = 4e-3, type = float,
                        help="Initial learning rate.")
    parser.add_argument('--batch_size', default = 200, type = int,
                        help="Training batch size.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument('--report', default = 10, type = int,
                        help = "The interval of training rounds to report.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--decay', type = float, default = 0.99,
                        help="The decay factor of the moving average.")
    parser.add_argument('--diffSig',action="store_true",dest = "diff",
                        help="If the input chunks are diffrential signal.")
    args = parser.parse_args(sys.argv[1:])
    args.seq = None
    args.seq_len = None
    if not os.path.isdir(args.model_folder):
        os.mkdir(args.model_folder)
    main(args)



"""
@author: Haotian Teng
"""
import torch
from torch import nn
from functools import partial
from typing import List


Conv1dk1 = partial(nn.Conv1d,kernel_size = 1)
RevConv1dk1 = partial(nn.ConvTranspose1d, kernel_size = 1)

class AttentionNormalize(nn.Module):
    ##TODO: this module makes the model hard to converge, need to modify.
    def __init__(self, 
                 in_channels: int, 
                 out_channels: int = 5, 
                 activation: nn.Module = nn.SiLU):
        super().__init__()
        hidden_num = out_channels
        self.hidden_num = hidden_num
        self.self_map = nn.Conv1d(in_channels,
                                  hidden_num,
                                  stride = 1,
                                  kernel_size = 1)
        self.ReLU = nn.ReLU()
        self.wmf = torch.nn.Linear(in_channels,hidden_num)
        self.wmb = torch.nn.Linear(in_channels,hidden_num)
        self.wk = torch.nn.Linear(hidden_num,hidden_num)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.self_map(x) #[N,C,L]
        out = self.wmf(torch.mean(x,dim=2,keepdim=False))#[N,C]
        out = out.unsqueeze(dim = 1)*self.wmb(x.permute(0,2,1)) + x0.permute(0,2,1) #[N,L,C]
        k = self.wk(out)
        scale = torch.mean(torch.sum(k*out,dim = 2),dim = 1) #[N]
        scale = self.ReLU(scale)
        scale = (0.2+scale)/torch.sqrt(1.2+scale)/torch.sqrt(torch.tensor(self.hidden_num))
        out = out*scale[:,None,None]
        # out = self.norm(out)#[N,L,C]
        return out.permute(0,2,1) #[N,C,L]

class Res1d(nn.Module):
    def __init__(self, 
                 in_channels: int, 
                 out_channels: int, 
                 kernel_size: int = 3, 
                 stride: int = 1,
                 activation: nn.Module = nn.SiLU,
                 batch_norm: nn.Module = nn.BatchNorm1d):
        super().__init__()
        self.self_map = nn.Conv1d(in_channels,
                                  out_channels,
                                  stride = stride,
                                  kernel_size = stride)
        
        self.conv1 = Conv1dk1(in_channels, out_channels,stride=1)
        self.bn1 = batch_norm(out_channels)
        self.conv2 = nn.Conv1d(out_channels, 
                               out_channels, 
                               kernel_size = kernel_size,
                               stride = stride,
                               padding = (kernel_size-stride)//2)
        self.bn2 = batch_norm(out_channels)
        self.activation = activation(inplace = True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.self_map(x)
        out = self.conv1(x)
        if isinstance(self.bn1,nn.LayerNorm):
            out = self.bn1(out.permute(0,2,1)).permute(0,2,1)
        else:
            out = self.bn1(out)
        out = self.activation(out)
        out = self.conv2(out)
        if isinstance(self.bn2,nn.LayerNorm):
            out = self.bn2(out.permute(0,2,1)).permute(0,2,1)
        else:
            out = self.bn2(out)
        out = self.activation(out)
        out += x0
        return out

class RevRes1d(nn.Module):
    def __init__(self, 
                 in_channels: int, 
                 out_channels: int, 
                 kernel_size: int = 3, 
                 stride: int = 1,
                 activation: nn.Module = nn.SiLU,
                 batch_norm: nn.Module = nn.BatchNorm1d):
        super().__init__()
        self.self_map = nn.ConvTranspose1d(in_channels,
                                           out_channels,
                                           kernel_size = stride,
                                           stride = stride)
        self.conv1 = Conv1dk1(in_channels, out_channels,stride=1)
        self.bn1 = batch_norm(out_channels)
        self.conv2 = nn.ConvTranspose1d(out_channels, 
                                        out_channels, 
                                        kernel_size = kernel_size,
                                        stride = stride,
                                        padding = (kernel_size-stride)//2)
        self.bn2 = batch_norm(out_channels)
        self.activation = activation(inplace = True)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x0 = self.self_map(x)
        
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.activation(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.activation(out)
        
        out += x0
        return out
    
class BidirectionalRNN(nn.Module):
    def __init__(self,
                 input_size: int,
                 hidden_size: int,
                 num_layers: int,
                 cell: nn.Module = nn.LSTM):
        super(BidirectionalRNN,self).__init__()
        self.rnn = cell(input_size,hidden_size,num_layers,bidirectional = True)
        self.num_layers = num_layers
        self.hidden_size = hidden_size
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        output,(hn,cn) = self.rnn(x)
        return output

class Permute(nn.Module):
    def __init__(self,perm:List):
        super().__init__()
        self.perm = perm
    def forward(self,x:torch.Tensor):
        return x.permute(self.perm)



"""
Created on Mon Apr 19 13:59:01 2021

@author: Haotian Teng
"""
from numpy import ndarray
import os
import h5py
import itertools
import numpy as np
from tqdm import tqdm
from typing import List
import contextlib
from typing import Dict
from scipy.signal import find_peaks

def arr2seq(arr:ndarray, base_dict):
    return ''.join([base_dict[x] for x in arr])

def raw2seq(raw:ndarray, blank_symbol:int = 0)->List[ndarray]:
    """
    Transfer raw ctc sequence by removing the consecutive repeat and blank symbol.

    Parameters
    ----------
    raw : ndarray
        The [N,C,L] raw ctc sequence, where N is the batch size and L is the 
        sequence length.
    blank_symbol : int, optional
        The index of the blank symbol. The default is 0.

    Returns
    -------
    List[ndarray]
        Length N list contains Transfered sequences.

    """
    mask = np.diff(raw).astype(bool)
    mask = np.insert(mask,0,True,axis = 1)
    moves = np.logical_and(raw!=blank_symbol,mask)
    seqs = np.ma.array(raw,mask = np.logical_not(mask)).tolist(None)
    out_seqs = [[x-1 for x in seq if x is not None and x!=blank_symbol] for seq in seqs]
    return out_seqs,moves

def list2string(input_v, base_type):
    if base_type == 0:
        base_dict = {0: 'A', 1: 'C', 2: 'G', 3: 'T', 4: 'b'}
    if base_type == 1:
        base_dict = {0: 'A', 1: 'C', 2: 'G', 3: 'T'}
    if base_type == 2:
        base_dict = {0: 'A', 1: 'C', 2: 'G', 3: 'T', 4: 'M'}
    if isinstance(base_type,dict):
        base_dict = base_type
    return "".join([base_dict[item] for item in input_v])


def string2list(input_v, base_type):
    if base_type == 0:
        base_dict = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
    if isinstance(base_type,dict):
        base_dict = base_type
    result = list()
    for item in input_v:
        result.append(base_dict[item])
    return result

def findall(string,char):
    return [idx for idx,c in enumerate(string) if c==char]
        
def seq2kmers(seq:str,k:int = 3):
    return [seq[x:x+k] for x in np.arange(len(seq) - k + 1)]

def kmer2int(kmer,base_order = ['A','C','G','T','M']):
    base = len(base_order)
    s = 0
    place = 0
    for b in kmer[::-1]:
        s += base_order.index(b)*base**place
        place += 1
    return s

def kmers2array(kmers,base_order = ['A','C','G','T','M']):
    return [kmer2int(x) for x in kmers]

def fast5_iter_old(fast5_dir,mode = 'r'):
    for (dirpath, dirnames, filenames) in os.walk(fast5_dir+'/'):
        for filename in filenames:
            if not filename.endswith('fast5'):
                continue
            abs_path = os.path.join(dirpath,filename)
            root = h5py.File(abs_path,mode = mode)
            read_h = list(root['/Raw/Reads'].values())[0]
            if 'Signal_Old' in read_h:
                signal = np.asarray(read_h[('Signal_Old')],dtype = np.float32)
            else:
                signal = np.asarray(read_h[('Signal')],dtype = np.float32)
            read_id = read_h.attrs['read_id']
            if type(read_id) != type('a'):
                read_id = read_id.decode("utf-8")
            yield read_h,signal,root,read_id
           
class NullContextManager(object):
    def __init__(self, dummy_resource=None):
        self.dummy_resource = dummy_resource
    def __enter__(self):
        return self.dummy_resource
    def __exit__(self, *args):
        pass

def fast5_shallow_iter(fast5_dir,mode = 'r',tqdm_bar = False):
    fail_count = 0
    with tqdm() if tqdm_bar else NullContextManager() as t:
        for (dirpath, dirnames, filenames) in os.walk(fast5_dir+'/'):
            for filename in filenames:
                if not filename.endswith('fast5'):
                    continue
                abs_path = os.path.join(dirpath,filename)
                try:
                    root = h5py.File(abs_path,mode = mode)
                    if tqdm_bar:
                        t.postfix = "File: %s, failed: %d"%(abs_path,fail_count)
                        t.update()
                    yield root,abs_path
                    root.close()
                except Exception as e:
                    print("Reading %s failed due to %s."%(abs_path,e))
                    fail_count += 1
                    continue

def fast5_iter(fast5_dir,mode = 'r',tqdm_bar = False):
    fail_count = 0
    read_id = ""
    with tqdm() if tqdm_bar else NullContextManager() as t:
        for (dirpath, dirnames, filenames) in os.walk(fast5_dir+'/'):
            for filename in filenames:
                if not filename.endswith('fast5'):
                    continue
                abs_path = os.path.join(dirpath,filename)
                try:
                    root = h5py.File(abs_path,mode = mode)
                except Exception as e:
                    print("Reading %s failed due to %s."%(abs_path,e))
                    fail_count += 1
                    if tqdm_bar:
                        t.postfix = "File: %s, Read: %s, failed: %d"%(abs_path,read_id,fail_count)
                        t.update()
                    continue
                if 'Raw' in root:
                    read_h = list(root['/Raw/Reads'].values())[0]
                    if 'Signal_Old' in read_h:
                        signal = np.asarray(read_h[('Signal_Old')],dtype = np.float32)
                    else:
                        signal = np.asarray(read_h[('Signal')],dtype = np.float32)
                    read_id = read_h.attrs['read_id']
                    if type(read_id) != type('a'):
                        read_id = read_id.decode("utf-8")
                    if tqdm_bar:
                        t.postfix = "File: %s, Read: %s, failed: %d"%(abs_path,read_id,fail_count)
                        t.update()
                    yield root,signal,abs_path,read_id
                else:
                    for read_id in root:
                        try:
                            read_h = root[read_id]
                            signal_h = read_h['Raw']
                            signal = np.asarray(signal_h[('Signal')],dtype = np.float32)
                            read_id = signal_h.attrs['read_id']
                            if tqdm_bar:
                                t.postfix = "File: %s, Read: %s, failed: %d"%(abs_path,read_id,fail_count)
                                t.update()
                            yield read_h,signal,abs_path,read_id.decode("utf-8")
                        except Exception as e:
                            print("Reading %s failed due to %s."%(read_id,e))
                            fail_count += 1
                            if tqdm_bar:
                                t.postfix = "File: %s, Read: %s, failed: %d"%(abs_path,read_id,fail_count)
                                t.update()
                            continue
                root.close()
            
def med_mad(x, factor=1.4826):
    """
    Calculate signal median and median absolute deviation
    Same method used in Bonito Basecaller.
    """
    med = np.median(x)
    mad = np.median(np.absolute(x - med)) * factor
    return med, mad


def combine_normalization(signal:np.array, kmer_seqs:np.array)-> np.array:
    """
    Normalize the signal given the kmer sequence, the normalization is done in
    a dwell-aware way, normalization first is calculated inside each dwell, and
    then calculated among the dwells.

    Parameters
    ----------
    seignal : np.array with shape [L]
        The unnormalized signal.
    kmer_seqs : np.array with shape [L]
        The array with same shape as the signal gives the hidden 
        representation (kmer) of the signal.

    Returns
    -------
    norm_sig: np.array with shape [L]
        Give the normalized signal.

    """
    combined = list(zip(signal,kmer_seqs))
    grouped = [list(g) for k,g in itertools.groupby(combined,key = lambda x:x[1])]
    grouped = [[x[0] for x in dwell] for dwell in grouped ]
    dwell_mean = [np.mean(dwell) for dwell in grouped]
    dwell_var = [np.var(dwell) for dwell in grouped]
    return (signal - np.mean(dwell_mean))/np.sqrt(np.mean(dwell_var)+np.var(dwell_mean))

def dwell_normalization(signal:np.array, kmer_seqs:np.array)-> np.array:
    """
    Normalize the signal given the kmer sequence, the normalization is done in
    a dwell-aware way, normalization first is calculated inside each dwell, and
    then calculated among the dwells.

    Parameters
    ----------
    seignal : np.array with shape [L]
        The unnormalized signal.
    kmer_seqs : np.array with shape [L]
        The array with same shape as the signal gives the hidden 
        representation (kmer) of the signal.

    Returns
    -------
    norm_sig: np.array with shape [L]
        Give the normalized signal.

    """
    combined = list(zip(signal,kmer_seqs))
    grouped = [list(g) for k,g in itertools.groupby(combined,key = lambda x:x[1])]
    grouped = [[x[0] for x in dwell] for dwell in grouped ]
    dwell_mean = [np.mean(dwell) for dwell in grouped if len(dwell)>1]
    dwell_var = [np.var(dwell) for dwell in grouped if len(dwell)>1]
    return (signal - np.mean(dwell_mean))/np.sqrt(np.mean(dwell_var))

def med_normalization(signal:np.array) -> np.array:
    med, mad = med_mad(signal)
    return (signal - med) / mad

def length_mask(length:np.array, max_length:int):
    b = length.shape[0]
    if len(length.shape) == 1:
        length = length[:,None]
    return np.tile(np.arange(max_length),b).reshape(b,max_length)<length
        

def norm_by_noisiest_section(signal, samples=100, threshold=6.0,offset = 0.0):
    """
    Normalise using the medmad from the longest continuous region where the
    noise is above some threshold relative to the std of the full signal.This
    function is borrowed from Bonito Basecaller.

    Parameters
    ----------
    signal : 1-D np.array
        A 1D numpy array contain signal from one read.
    samples : Int, optional
        The window size. The default is 100.
    threshold : Float, optional
        Signal with std over threshold are used to find the peaks. The default is 6.0.
    offset : Float, optional
        The offset to apply to the signal. The default is 0.0.

    Returns
    -------
    normalized_signal: 1-D array
        The normalized signal.
    med : Float
        The median value.
    mad : Float
        The MAD value estimate the deviation of the signal.

    """
    threshold = signal.std() / threshold
    noise = np.ones(signal.shape)

    for idx in np.arange(signal.shape[0] // samples):
        window = slice(idx * samples, (idx + 1) * samples)
        noise[window] = np.where(signal[window].std() > threshold, 1, 0)

    # start and end low for peak finding
    noise[0] = 0; noise[-1] = 0
    peaks, info = find_peaks(noise, width=(None, None))

    if len(peaks):
        widest = np.argmax(info['widths'])
        med, mad = med_mad(signal[info['left_bases'][widest]: info['right_bases'][widest]])
    else:
        med, mad = med_mad(signal)
    return (signal - med + offset) / mad, med, mad

def diff_norm_by_noisiest_section(signal, samples=100, threshold=6.0,offset = 0.0):
    """
    Calculate the difference of the signal S[1:]-S[:-1], scale the
    diff signal with the MAD value that extracted with the same method as
    norm_by_noisiest_section function.

    Parameters
    ----------
    signal : 1-D np.array
        A 1D numpy array contain signal from one read.
    samples : Int, optional
        The window size. The default is 100.
    threshold : Float, optional
        Signal with std over threshold are used to find the peaks. The default is 6.0.
    offset : Float, optional
        The offset to apply to the signal. The default is 0.0.

    Returns
    -------
    normalized_signal: 1-D array
        The differentiable normalized signal.
    med : Float
        The median value.
    mad : Float
        The MAD value estimate the deviation of the signal.

    """
    threshold = signal.std() / threshold
    noise = np.ones(signal.shape)

    for idx in np.arange(signal.shape[0] // samples):
        window = slice(idx * samples, (idx + 1) * samples)
        noise[window] = np.where(signal[window].std() > threshold, 1, 0)

    # start and end low for peak finding
    noise[0] = 0; noise[-1] = 0
    peaks, info = find_peaks(noise, width=(None, None))
    diff_signal = signal[1:] - signal[:-1]
    if len(peaks):
        widest = np.argmax(info['widths'])
        med, mad = med_mad(signal[info['left_bases'][widest]: info['right_bases'][widest]])
    else:
        med, mad = med_mad(signal)
    return (diff_signal + offset) / mad, med, mad

def diff_norm_fixing_deviation(signal, deviation = 100.0, offset = 0.0):
    """
    Return the differential signal by fixed value: (x_i-x_{i-1})/deviation

    Parameters
    ----------
    signal : 1-D np.array
        A 1D numpy array contain signal from one read.
    deviation: Float, optiona
        Scale the signal by 1/deviation.
    offset : Float, optional
        The offset to apply to the signal. The default is 0.0.

    Returns
    -------
    normalized_signal: 1-D array
        The differentiable normalized signal.
    med : Float
        The median value.
    mad : Float
        The MAD value estimate the deviation of the signal.

    """
    diff_signal = (signal[1:] - signal[:-1])/deviation
    med, mad = med_mad(diff_signal)
    return diff_signal + offset, med, mad

class Methylation_DP_Aligner(object):
    def __init__(self,
                 pxy:int = 2, 
                 pgopen: int = 4,
                 pgap:int = 3,
                 edge_gap:float = -0.1,
                 base_alternation:Dict = {}):
        """
        A sequence aligner that uses dynamic programming

        Parameters
        ----------
        pxy : int, optional
            The mismatch penalty score, default is 2.
        pgopen: int optional
            The penalty score for opening a gap, default is 3.5.
        pgap : int, optional
            The gap penalty, default is 3.
        edge_gap: int, optional
            The penalty reduce for the gap happened in edge, 
        base_alternation: dict, optional
            The dictionary gives the base alternation, default is None.
        
        """
        self.pxy = pxy
        self.pgopen = pgopen
        self.pgap = pgap
        self.edge_gap = edge_gap
        self.base_alt = base_alternation
        
    def _equal_base(self,a,b):
        if a in self.base_alt.keys():
            a = self.base_alt[a]
        if b in self.base_alt.keys():
            b = self.base_alt[b]
        return a==b
    
    def align(self, x:str, y:str):
        i,j = 0,0
        m,n = len(x),len(y)
         
        # table for storing optimal substructure answers
        dp = np.zeros([m+1,n+1], dtype = float) #The score matrix
        d = np.zeros([m+1,n+1], dtype = int) #The direction matrix
        
        # initialising the table
        dp[:,0] = self.pgap * np.arange(m+1)
        dp[0,:] = self.pgap * np.arange(n+1) 
        d[:,0] = 1
        d[0,:] = 2
        # calculating the minimum penalty
        i = 1
        while i <= m:
            j = 1
            while j <= n:
                curr = [dp[i - 1][j - 1] + (0 if self._equal_base(x[i-1], y[j-1]) else self.pxy),
                        dp[i - 1][j] + (self.pgap if d[i-1][j] == 1 else self.pgopen)+(self.edge_gap if j == n else 0),
                        dp[i][j - 1] + (self.pgap if d[i][j-1] == 2 else self.pgopen)+(self.edge_gap if i == m else 0)]
                d[i][j] = max(np.where(curr == np.min(curr))[0])
                dp[i][j] = curr[d[i][j]]
                    
                j += 1
            i += 1
         
        # Reconstructing the solution
        l = n + m   # maximum possible length
        i = m
        j = n
        xpos = l
        ypos = l
     
        # Final answers for the respective strings
        xans = np.zeros(l+1, dtype=int)
        yans = np.zeros(l+1, dtype=int)
         
     
        while not (i == 0 or j == 0):
            #print(f"i: {i}, j: {j}")
            if d[i][j] == 0:       
                xans[xpos] = ord(x[i - 1])
                yans[ypos] = ord(y[j - 1])
                xpos -= 1
                ypos -= 1
                i -= 1
                j -= 1
             
            elif d[i][j] == 1:
                xans[xpos] = ord(x[i - 1])
                yans[ypos] = ord('_')
                xpos -= 1
                ypos -= 1
                i -= 1
             
            elif d[i][j] == 2:
                xans[xpos] = ord('_')
                yans[ypos] = ord(y[j - 1])
                xpos -= 1
                ypos -= 1
                j -= 1
             
     
        while xpos > 0:
            if i > 0:
                i -= 1
                xans[xpos] = ord(x[i])
                xpos -= 1
            else:
                xans[xpos] = ord('_')
                xpos -= 1
         
        while ypos > 0:
            if j > 0:
                j -= 1
                yans[ypos] = ord(y[j])
                ypos -= 1
            else:
                yans[ypos] = ord('_')
                ypos -= 1
        mask = np.logical_not((xans==yans)*(xans == ord('_')))
        seq_x = ''.join([chr(x) for x in xans[mask]][1:])
        seq_y = ''.join([chr(x) for x in yans[mask]][1:])
        return seq_x,seq_y
    
    def merge(self,seq_x,ref_algn):
        rs_x = seq_x.rstrip('_')
        ref_algn = ref_algn[:len(rs_x)] #Trim the right side of the reference
        m_seq = [(x if x!="_" else y) for x,y in zip(seq_x,ref_algn) if y!="_"]
        return ''.join(m_seq)
    
if __name__ == "__main__":
    aligner = Methylation_DP_Aligner()
    x = "AAAGAATCA"
    y = "AAAAGAATTCACA"
    x_,y_ = aligner.align(x,y)
    print(aligner.merge(x_,y_))


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# Copyright 2017 The Chiron Authors. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Created on Mon Mar 27 14:04:57 2017
# from rnn import rnn_layers
from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import time
import argparse

import tensorflow as tf
import xron.xron_model as model
from xron.xron_input import read_raw_data_sets
from xron.xron_input import read_cache_dataset
from tensorflow.python.ops import variables
from six.moves import range
DEFAULT_OFFSET = 10
tf.logging.set_verbosity(tf.logging.ERROR)
def save_hyper_parameter():
    """
    TODO: Function to save the hyper parameter.
    """
def compile_train_graph(config,hp):
    class net:
        pass
    net.training = tf.placeholder(tf.bool)
    net.global_step = tf.get_variable('global_step', trainable=False, shape=(),
                                  dtype=tf.int32,
                                  initializer=tf.zeros_initializer())
    net.x = tf.placeholder(tf.float32, shape=[hp.batch_size, hp.sequence_len])
    net.seq_length = tf.placeholder(tf.int32, shape=[hp.batch_size])
    net.y_indexs = tf.placeholder(tf.int64)
    net.y_values = tf.placeholder(tf.int32)
    net.y_shape = tf.placeholder(tf.int64)
    net.y = tf.SparseTensor(net.y_indexs, net.y_values, net.y_shape)
    net.logits, net.ratio = model.inference(net.x, net.seq_length, net.training,hp.sequence_len,configure = config)
    if 'fl_gamma' in config.keys():
        net.ctc_loss = model.loss(net.logits, net.seq_length, net.y, fl_gamma = config['fl_gamma'])
    else:
        net.ctc_loss = model.loss(net.logits, net.seq_length, net.y)
    net.opt = model.train_opt(hp.step_rate,
                          hp.max_steps, 
                          global_step=net.global_step,
                          opt_name = config['opt_method'])
    if hp.gradient_clip is None:
        net.step = net.opt.minimize(net.ctc_loss,global_step = net.global_step)
    else:
        net.gradients, net.variables = zip(*net.opt.compute_gradients(net.ctc_loss))
        net.gradients = [None if gradient is None else tf.clip_by_norm(gradient, hp.gradient_clip) for gradient in net.gradients]
        net.step = net.opt.apply_gradients(zip(net.gradients, net.variables),global_step = net.global_step)
    net.error,net.errors,net.y_ = model.prediction(net.logits, net.seq_length, net.y)
    net.init = tf.global_variables_initializer()
    net.variable_to_restore=set(variables._all_saveable_objects()+tf.moving_average_variables())
    net.saver = tf.train.Saver(var_list=net.variable_to_restore, 
                               save_relative_paths=True)
    net.summary = tf.summary.merge_all()
    return net

def train():
    default_config = os.path.join(FLAGS.log_dir,FLAGS.model_name,'model.json')
    if FLAGS.retrain:
        if os.path.isfile(default_config):
            config_file = default_config
        else:
            raise ValueError("Model Json file has not been found in model log directory")
    else:
        config_file = FLAGS.configure   
    config = model.read_config(config_file)
    print("Begin training using following setting:")
    with open(os.path.join(FLAGS.log_dir,FLAGS.model_name,'train_config'),'w+') as log_f:
        for pro in dir(FLAGS):
            if not pro.startswith('_'):
                print("%s:%s"%(pro,getattr(FLAGS,pro)))
                log_f.write("%s:%s\n"%(pro,getattr(FLAGS,pro)))
    net = compile_train_graph(config,FLAGS)
    sess = tf.Session(config=tf.ConfigProto(inter_op_parallelism_threads=FLAGS.threads,
                                            intra_op_parallelism_threads=FLAGS.threads,
                                            allow_soft_placement=True))
    if FLAGS.retrain == False:
        sess.run(net.init)
        print("Model init finished, begin loading data. \n")
    else:
        net.saver.restore(sess, tf.train.latest_checkpoint(
            FLAGS.log_dir + FLAGS.model_name))
        print("Model loaded finished, begin loading data. \n")
    summary_writer = tf.summary.FileWriter(
        FLAGS.log_dir + FLAGS.model_name + '/summary/', sess.graph)
    model.save_model(default_config,config)
    train_ds,valid_ds = generate_train_valid_datasets(initial_offset = DEFAULT_OFFSET)
    start = time.time()
    resample_n = 0
    for i in range(FLAGS.max_steps):
        if FLAGS.resample_after_epoch == 0:
            pass
        elif train_ds.epochs_completed >= FLAGS.resample_after_epoch:
            train_ds,valid_ds = generate_train_valid_datasets(initial_offset = resample_n*FLAGS.offset_increment + DEFAULT_OFFSET)
        batch_x, seq_len, batch_y = train_ds.next_batch(FLAGS.batch_size)
        indxs, values, shape = batch_y
        feed_dict = {net.x: batch_x, net.seq_length: seq_len / net.ratio, net.y_indexs: indxs,
                     net.y_values: values, net.y_shape: shape,
                     net.training: True}
        loss_val, _ = sess.run([net.ctc_loss, net.step], feed_dict=feed_dict)
        if i % 10 == 0:
            global_step_val = tf.train.global_step(sess, net.global_step)
            valid_x, valid_len, valid_y = valid_ds.next_batch(FLAGS.batch_size)
            indxs, values, shape = valid_y
            feed_dict = {net.x: valid_x, net.seq_length: valid_len / net.ratio,
                         net.y_indexs: indxs, net.y_values: values, net.y_shape: shape,
                         net.training: True}
            error_val = sess.run(net.error, feed_dict=feed_dict)
#            x_val,errors_val,y_predict,y = sess.run([x,errors,y_,y],feed_dict = feed_dict)
#            predict_seq,_ = sparse2dense([y_predict,0])
#            true_seq,_ = sparse2dense([[y],0])
            end = time.time()
            print(
            "Step %d/%d Epoch %d, batch number %d, train_loss: %5.3f validate_edit_distance: %5.3f Elapsed Time/step: %5.3f" \
            % (i, FLAGS.max_steps, train_ds.epochs_completed,
               train_ds.index_in_epoch, loss_val, error_val,
               (end - start) / (i + 1)))
            net.saver.save(sess, FLAGS.log_dir + FLAGS.model_name + '/model.ckpt',
                       global_step=global_step_val)
            summary_str = sess.run(net.summary, feed_dict=feed_dict)
            summary_writer.add_summary(summary_str, global_step=global_step_val)
            summary_writer.flush()
    global_step_val = tf.train.global_step(sess, net.global_step)
    print("Model %s saved." % (FLAGS.log_dir + FLAGS.model_name))
    print("Reads number %d" % (train_ds.reads_n))
    net.saver.save(sess, FLAGS.log_dir + FLAGS.model_name + '/final.ckpt',
               global_step=global_step_val)
    
def generate_train_valid_datasets(initial_offset = 10):
    if FLAGS.read_cache:
        train_ds = read_cache_dataset(FLAGS.train_cache)
        if FLAGS.validation is not None:
            valid_ds = read_cache_dataset(FLAGS.valid_cache)
        else:
            valid_ds = train_ds
        if train_ds.event.shape[1]!=FLAGS.sequence_len:
            raise ValueError("The event length of training cached dataset %d is inconsistent with given sequene_len %d"%(train_ds.event.shape()[1],FLAGS.sequence_len))
        if valid_ds.event.shape[1]!=FLAGS.sequence_len:
            raise ValueError("The event length of training cached dataset %d is inconsistent with given sequene_len %d"%(valid_ds.event.shape()[1],FLAGS.sequence_len))
        return train_ds,valid_ds
    sys.stdout.write("Begin reading training dataset.\n")
    train_ds = read_raw_data_sets(FLAGS.data_dir,
                                  FLAGS.train_cache,
                                  FLAGS.sequence_len, 
                                  k_mer=FLAGS.k_mer,
                                  max_segments_num=FLAGS.segments_num,
                                  skip_start = initial_offset)
    sys.stdout.write("Begin reading validation dataset.\n")
    if FLAGS.validation is not None:
        valid_ds = read_raw_data_sets(FLAGS.validation,
                                      FLAGS.valid_cache,
                                      FLAGS.sequence_len, 
                                      k_mer=FLAGS.k_mer,
                                      max_segments_num=FLAGS.segments_num)
    else:
        valid_ds = train_ds
    return train_ds,valid_ds
def run(args):
    global FLAGS
    FLAGS = args
    if FLAGS.train_cache is None:
        FLAGS.train_cache = os.path.join(FLAGS.log_dir,FLAGS.model_name) + '/train_cache.hdf5'
    if (FLAGS.valid_cache is None) and (FLAGS.validation is not None):
        FLAGS.valid_cache = os.path.join(FLAGS.log_dir,FLAGS.model_name) + '/valid_cache.hdf5'
    FLAGS.data_dir = FLAGS.data_dir + os.path.sep
    FLAGS.log_dir = FLAGS.log_dir + os.path.sep
    train()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training model with tfrecord file')
    parser.add_argument('-i', '--data_dir', required = True,
                        help="Directory that store the tfrecord files.")
    parser.add_argument('-o', '--log_dir', required = True  ,
                        help="log directory that store the training model.")
    parser.add_argument('-m', '--model_name', required = True,
                        help='model_name')
    parser.add_argument('-v', '--validation', default = None, 
                        help="validation data folder, default is None, which use the train dataset.")
    parser.add_argument('--train_cache', default=None, help="Cache file for training dataset.")
    parser.add_argument('--valid_cache', default=None, help="Cache file for validation dataset.")
    parser.add_argument('-s', '--sequence_len', type=int, default=400,
                        help='the length of sequence')
    parser.add_argument('-b', '--batch_size', type=int, default=300,
                        help='Batch size')
    parser.add_argument('-t', '--step_rate', type=float, default=4e-3,
                        help='Step rate')
    parser.add_argument('-x', '--max_steps', type=int, default=10000,
                        help='Maximum step')
    parser.add_argument('-n', '--segments_num', type = int, default = None,
                        help='Maximum number of segments read into the training queue, default(None) read all segments.')
    parser.add_argument('--configure', default = None,
                        help="Model structure configure json file.")
    parser.add_argument('-k', '--k_mer', default=1, help='Output k-mer size')
    parser.add_argument('--resample_after_epoch',
                        type = int,
                        default = 0, 
                        help='Resample the reads data every n epoches, with an increasing initial offset.')
    parser.add_argument('--threads',
                        type = int,
                        default = 0, 
                        help='The threads that available, if 0 use all threads that can be found.')
    parser.add_argument('--offset_increment',
                        type = int,
                        default = 3,
                        help='The increament of initial offset if the resample_after_epoch has been set.')
    parser.add_argument('--gradient_clip',
                        type = float,
                        default = None,
                        help = 'Clip the gradient by the gradient_clip x normalization, a good estimate is 5.')
    parser.add_argument('--retrain', dest='retrain', action='store_true',
                        help='Set retrain to true')
    parser.add_argument('--read_cache',dest='read_cache',action='store_true',
                        help="Read from cached hdf5 file.")
    parser.set_defaults(retrain=False)
    args = parser.parse_args(sys.argv[1:])
    run(args)



"""
Created on Mon May  9 06:15:29 2022

@author: Haotian Teng
"""
import os
import sys
import time
import toml
import torch
import argparse
import itertools
import numpy as np
from matplotlib import pyplot as plt
from xron.nrhmm.hmm import GaussianEmissions, RHMM
from xron.nrhmm.hmm_input import Kmer2Transition, Kmer_Dataset
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader
from xron.xron_train_base import DeviceDataLoader





if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Evaluating using the RHMM model')
    parser.add_argument("-i","--input", type = str, required = True,
                        help = "Data folder contains the chunk, kmer sequence.")
    parser.add_argument('-o', '--output', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument("-b","--batch_size", type = int, default = 50,
                        help = "The batch size to train.")
    parser.add_argument("--lr",type = float, default = 1e-2,
                        help = "The initial training learning rate.")
    parser.add_argument("--report",type = int, default = 10,
                        help = "Report the loss and save the model every report cycle.")
    parser.add_argument("--certain_methylation",action = "store_false", 
                        dest = "kmer_replacement", 
                        help = "If we are sure about the methylation state.")
    parser.add_argument("--optimizer",type = str, default = "Adam",
                        help = "The optimizer used to train the model.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument("--device",type = str, default = "cuda",
                        help = "The device used to train the model.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--moving_average', type = float, default = 0.0,
                         help="The factor of moving average, 0 means no delay.")
    parser.add_argument('--trainable_bases', type = str, default = None,
                        help="A magic string AB!CD or AB that gives the trainable bases and NOT trainalbe bases, for example MC means trains on kmer that must contains M and C, A!M means trains on kmer that must contains A but not contains M.")
    args = parser.parse_args(sys.argv[1:])
    os.makedirs(args.model_folder,exist_ok=True)
    train(args)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  3 11:47:48 2022
This script is slower than single thread index.py, probably because File IO is the bottleneck.
use index.py instead.
@author: heavens
"""
import os
import h5py
import argparse
from xron.utils.seq_op import fast5_shallow_iter
from xron.utils.fastIO import read_fastq
import multiprocessing as mp
from time import sleep 
from multiprocessing import Process
from multiprocessing import Manager
from multiprocessing import Queue
from tqdm import tqdm
import queue #In Python 2.7, this would be Queue

DESIRED_TIMEOUT = 1
MAX_result_dict_SIZE = 1000#This is the number of instances in the queue, the number should be set in the way that the max queue byte size smaller than the page capacity.
def worker(job_queue,fastqs,result_queue,log):
    while True:
        try:
            i = job_queue.get(timeout = DESIRED_TIMEOUT)
        except queue.Empty:
            return
        #Do the job here
        try:
            with h5py.File(i,mode = 'r') as root:
                read_ids = list(root.keys())
                for id in read_ids:
                    if id[5:] in fastqs['name']:
                        result_queue.put((id[5:],i))
            log['success'] += 1
        except Exception as e:
            print("Reading %s failed due to %s."%(i,e))
            log['fail'] += 1
            continue

def run(args):
    manager = Manager()  
    filequeue = manager.Queue()
    result_queue = manager.Queue()
    log = manager.dict()
    log['fail'] = 0
    log['success'] = 0
    
    max_threads_number = mp.cpu_count()-1 #1 thread is used for the main process
    all_proc = []
    fastq_records = read_fastq(args.fastq)
    fast5_records = {}
    file_number = 0
    print("Read in fast5 file list.")
    for (dirpath, dirnames, filenames) in os.walk(args.fast5+'/'):
        for filename in filenames:
            if not filename.endswith('fast5'):
                continue
            abs_path = os.path.join(dirpath,filename)
            filequeue.put(abs_path)
            file_number += 1
    print("Create read id to fast5 file mapping.")
    for i in range(max_threads_number if args.threads is None else args.threads):
        p = Process(target = worker, args = (filequeue,fastq_records,result_queue,log))
        all_proc.append(p)
        p.start()
    
    print("Indexing fastq files.")
    with open(args.fastq+'.index', 'w+') as f:
        with tqdm() as t:
            t.total = len(fastq_records['name'])
            while log['success'] + log['fail'] < file_number or not filequeue.empty() or not result_queue.empty():
                try:
                    #pop out result from the dict
                    result = result_queue.get()
                    if result[0] in fastq_records['name']:
                        f.write(result[0]+'\t'+result[1]+'\n')
                        fastq_records['name'].pop(fastq_records['name'].index(result[0]))
                        t.update()
                except KeyError:
                    sleep(0.1)
                    continue
    if len(fastq_records['name']) != 0:
        raise ValueError('%d fastq readid not found in fast5'%(len(fastq_records['name'])))       

    for p in all_proc:
        p.join()
 
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('--fast5', required = True, type=str, help='folder that contains fast5 output')
    parser.add_argument('--fastq', required = True, type=str, help='The merged fastq file')
    parser.add_argument('--threads', type=int, default = None, help='The number of threads used for indexing')
    args = parser.parse_args()
    run(args)


"""
Created on Thu Apr 28 18:03:07 2022

@author: Haotian Teng
"""
import torch
import itertools
import numpy as np
import timeit

def sparse_multiplication(transition,log_alpha):
    idxs = transition.indices()
    vals = transition.values().log()
    update_check = torch.zeros((batch_size,n_kmers),dtype = torch.bool)
    update = log_alpha[idxs[0],idxs[1]] +vals
    for i,u in enumerate(update):
        batch_i,kmer_i = idxs[0][i],idxs[2][i]
        if not update_check[batch_i,kmer_i]:
            curr_update = update[torch.logical_and(idxs[0]==batch_i,idxs[2] == kmer_i)]
            log_alpha[idxs[0][i],idxs[2][i]] = torch.logsumexp(curr_update,dim = 0)
            update_check[batch_i,kmer_i] = True
    log_alpha[torch.logical_not(update_check)] += np.log(1e-6)
    return log_alpha
        
def gpu_multiplication(transition,log_alpha):
    dense_t = transition.to_dense()
    return log_domain_matmul(log_alpha, dense_t.log())

def base_multiplication(transition_base,indexing,log_alpha):
    n_kmer,n_base = indexing.shape
    alpha_shape = log_alpha.shape
    multiplication = log_alpha[:,indexing]
    return torch.logsumexp(transition_base.log() + multiplication,dim = -1)

def log_domain_sparse_matmul(A:torch.sparse_coo_tensor, log_B:torch.Tensor, dim:int = -1):
    """
    Do a sparse-dense tensor multiplication and reduced on the given dimension.

    Parameters
    ----------
    A : torch.sparse_coo_tensor
        A sparse tensor with shape mxnxp.
    log_B : torch.Tensor
        A dense tensor in the log domain with same shape, broadcast is supported on dense tensor.
    dim : int, optional
        The dimension to perform reduction on A. The default is -1.

    Returns
    -------
    A sparse tensor.

    """
    A = A.coalesce()
    idxs = A.indices()
    log_vals = A.values().log()
    shape_A = torch.tensor(A.shape)
    shape_B = torch.tensor(log_B.shape)
    n_dims_A = idxs.shape[0]
    n_dims_B = len(shape_B)
    assert n_dims_A == n_dims_B, "Tensor has different number of dimensions."
    assert torch.all(shape_A >= shape_B), "Broadcast only supported on dense tensor."
    idxs_B = idxs.clone()
    remain_dims = np.arange(n_dims_A)
    remain_dims = np.delete(remain_dims,dim)
    remain_idxs = list(zip(*[idxs.cpu()[x].tolist() for x in remain_dims]))
    idxs_B[torch.where(shape_B==1)] = 0
    update = log_B[tuple(idxs_B)] + log_vals
    key_func = lambda x: x[1]
    update = update.tolist()
    update = sorted(zip(update,remain_idxs),key = key_func)
    nested = [ (k,list(g)) for k,g in itertools.groupby(update,key = key_func)]
    nested_vals = [[y[0] for y in x[1]] for x in nested]
    nested_idxs = [x[0] for x in nested]
    max_cols = max([len(x) for x in nested_vals])
    padded = torch.tensor([x + [-np.inf]*(max_cols - len(x)) for x in nested_vals],device = A.device)
    return torch.sparse_coo_tensor(indices = list(zip(*nested_idxs)),
                                   values = torch.logsumexp(padded,dim = 1),
                                   size = shape_A[remain_dims].tolist(),
                                   device = A.device)


def log_domain_matmul(log_A, log_B):
    """
    log_A : m x n
    log_B : m x n x p or n x p
    output : m x p matrix
    
    Normally, a matrix multiplication
    computes out_{i,j} = sum_k A_{i,k} x B_{i,k,j}

	A log domain matrix multiplication
	computes out_{i,j} = logsumexp_k log_A_{i,k} + log_B_{i,k,j}
	"""
    dim_B = len(log_B.shape)
    log_A = log_A.unsqueeze(dim = 2)
    if dim_B == 2:
        log_B = log_B.unsqueeze(dim = 0)
    elementwise_sum = log_A + log_B
    out = torch.logsumexp(elementwise_sum, dim=1)
    return out

if __name__ == "__main__":
    sparsity = 0.0014
    n_kmers = 3125
    batch_size = 2
    n_base = 5
    n_elements = int(n_kmers*n_kmers*sparsity)
    start = np.random.randint(low = 0, high = n_kmers - 1, size = n_elements)
    end = np.random.randint(low = 0, high = n_kmers - 1, size = n_elements)
    batch_i = np.random.randint(low = 0, high = batch_size - 1, size = n_elements)
    transition = torch.sparse_coo_tensor(indices = (batch_i,start,end),values = [1.]*len(start),size = (batch_size,n_kmers,n_kmers)).coalesce()
    transition_cuda = transition.to("cuda")
    log_alpha = torch.rand((batch_size,n_kmers))
    log_alpha_cuda = log_alpha.to("cuda")
    result = log_domain_sparse_matmul(transition_cuda,log_alpha_cuda.unsqueeze(dim = 2),dim = 1)
    result_gpu = gpu_multiplication(transition_cuda,log_alpha_cuda)
    assert torch.all(torch.isclose(result.to_dense()[torch.logical_not(torch.isinf(result_gpu))], result_gpu[torch.logical_not(torch.isinf(result_gpu))]))
    # timeit.timeit(sparse_multiplication(transition, log_alpha),number = 10000)
    # timeit.timeit(gpu_multiplication(transition_cuda,log_alpha_cuda),number = 10000)
    transition_base = torch.rand((batch_size,n_kmers,n_base+1)).to("cuda")
    base_index = torch.randint(low = 0,high = n_kmers,size = (n_kmers,n_base + 1)).to("cuda")
    base_multiplication(transition_base,base_index,log_alpha_cuda)
    # timeit.timeit(base_multiplication(transition_base,base_index,log_alpha_cuda),number = 10000)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 15 16:47:56 2022

@author: haotian
"""
import h5py
import os
import numpy as np
from collections import defaultdict
from matplotlib import pyplot as plt
from tqdm import tqdm


def seq2kmer(seq:str,k:int = 5):
    return [seq[x:x+k] for x,_ in enumerate(seq[:-k+1])]

prefix = "/home/heavens/bridge_scratch/NA12878_RNA_IVT/guppy_train/fast5s"
kmer_dict_p1 = defaultdict(list)
kmer_dict_m1 = defaultdict(list)
k = 5
for f in tqdm(os.listdir(prefix)):
    if f.endswith("fast5"):
        with h5py.File(os.path.join(prefix,f),'r') as root:
            for read in root:
                try:
                    corrected = root[read]['Analyses/Segmentation_000/Reference_corrected']
                    signal = np.asarray(root[read]['Raw/Signal'])[::-1]
                    seq =np.asarray(corrected['ref_seq']).item().decode()
                    kmers = seq2kmer(seq,k = k)
                    ref_sig_idx = np.asarray(corrected['ref_sig_idx'])
                    for i in np.arange(len(kmers)):
                        c,c_n = ref_sig_idx[i+k//2],ref_sig_idx[i+k//2+1]
                        kmer = kmers[i]
                        kmer_dict_p1[kmer]+= list(signal[c:c_n])
                    for i in np.arange(len(kmers)):
                        c,c_n = ref_sig_idx[i+k//2-1],ref_sig_idx[i+k//2]
                        kmer = kmers[i]
                        kmer_dict_m1[kmer]+= list(signal[c:c_n])
                        
                except KeyError:
                    pass
    

std_m1 = []
std_p1 = []
for key,val in kmer_dict_m1.items():
    std_m1.append(np.std(val))
    std_p1.append(np.std(kmer_dict_p1[key]))


"""
@author: Haotian Teng
"""
import os 
import sys
import math
import torch
import argparse
import numpy as np
from torchvision import transforms
import torch.utils.data as data
from torch.utils.data.dataloader import DataLoader
from functools import partial
from xron.xron_model import CRNN, CONFIG
from xron.xron_input import Dataset, ToTensor, NumIndex
from xron.xron_train_base import Trainer, DeviceDataLoader, load_config

class SupervisedTrainer(Trainer):
    def __init__(self,
                 train_dataloader:DataLoader,
                 net:CRNN,
                 config:CONFIG,
                 device:str = None,
                 eval_dataloader:DataLoader = None,
                 logger = None):
        """

        Parameters
        ----------
        train_dataloader : DataLoader
            Training dataloader.
        net : CRNN
            A CRNN network instance.
        device: str
            The device used to train the model, can be 'cpu' or 'cuda'.
            Default is None, use cuda device if it's available.
        config: CONFIG
            A CONFIG class contains training configurations. Need to contain
            at least these parameters: keep_record, device and grad_norm.
        eval_dataloader : DataLoader, optional
            Evaluation dataloader, if None training dataloader will be used.
            The default is None.

        Returns
        -------
        None.

        """
        super().__init__(train_dataloader=train_dataloader,
                         nets = {"encoder":net},
                         config = config,
                         device = device,
                         eval_dataloader = eval_dataloader)
        self.net = net
        self.grad_norm = config.TRAIN['grad_norm']
        self.logger = logger
        
    def train(self,epoches,optimizer,save_cycle,valid_cycle,save_folder,scheduler = None):
        self.save_folder = save_folder
        self._save_config()
        initial_global_step = self.global_step
        for epoch_i in range(epoches):
            for i_batch, batch in enumerate(self.train_ds):
                if (i_batch+1)%save_cycle==0:
                    calculate_error = True
                else:
                    calculate_error = False
                loss,error = self.train_step(batch,get_error = calculate_error)
                if torch.isnan(loss):
                    print("NaN loss detected, skip this training step.")
                    continue
                optimizer.zero_grad()
                loss.backward()
                # torch.nn.utils.clip_grad_norm_(self.net.parameters(), 
                #                                max_norm=self.grad_norm)
                optimizer.step()
                self.losses.append(loss.item())
                self.global_step +=1
                if self.config.TRAIN['run_n_test']:
                    # run n test
                    if self.global_step - initial_global_step > self.config.TRAIN['run_n_test']:
                        return None
                if self.logger is not None:
                    self.logger.add_scalar('Loss/train', loss.item(), self.global_step)
                if (i_batch+1)%save_cycle==0:
                    self.save()
                    with torch.set_grad_enabled(False):
                        eval_i,valid_batch = next(enumerate(self.eval_ds))
                        valid_error,valid_perror = self.valid_step(valid_batch)
                        self.errors.append(valid_error)
                        if self.logger is not None:
                            self.logger.add_scalar('Error/train', error, self.global_step)
                            self.logger.add_scalar('Error/valid', valid_error, self.global_step)
                            self.logger.add_scalar('Error/valid_plain', valid_perror, self.global_step)
                        print("Epoch %d: Batch %d, loss %f, error %f valid_error %f, reducting_error %f"%(epoch_i, i_batch, loss,error, np.mean(valid_error),np.mean(valid_perror)))
                    self.save_loss()
                if (i_batch+1)%valid_cycle==0:
                    with torch.set_grad_enabled(False):
                        eval_i,valid_batch = next(enumerate(self.eval_ds))
                        valid_error,valid_perror = self.valid_step(valid_batch)
                        self.errors.append(valid_error)
                        print("Epoch %d: Batch %d, loss %f, error %f valid_error %f, reducting_error %f"%(epoch_i, i_batch, loss,error, np.mean(valid_error),np.mean(valid_perror)))
                    if scheduler is not None:
                        scheduler.step(valid_error)
                    if self.logger is not None:
                        self.logger.add_scalar('Error/train', error, self.global_step)
                        self.logger.add_scalar('Error/valid', valid_error, self.global_step)
                        self.logger.add_scalar('Error/valid_plain', valid_perror, self.global_step)
                
    def valid_step(self,batch):
        net = self.net
        net.eval()
        signal_batch = batch['signal']
        out = net.forward(signal_batch)
        seq = batch['seq']
        seq_len = batch['seq_len'].view(-1)
        error = None
        error = net.ctc_error(out,
                              seq,
                              seq_len,
                              alphabet = 'N' + self.config.CTC['alphabeta'],
                              beam_size = self.config.CTC['beam_size'],
                              beam_cut_threshold = self.config.CTC['beam_cut_threshold'])
        plain_error = net.ctc_error(out,
                                    seq,
                                    seq_len,
                                    alphabet = 'N' + self.config.CTC['alphabeta'],
                                    beam_size = self.config.CTC['beam_size'],
                                    beam_cut_threshold = self.config.CTC['beam_cut_threshold'],
                                    reduction = {'M':'A'})
        net.train()
        return error,plain_error

    def train_step(self,batch,get_error = False):
        net = self.net
        signal_batch = batch['signal']
        if torch.sum(torch.isnan(signal_batch)):
            print("Found NaN input signal.")
            return None,None
        out = net.forward(signal_batch)
        out_len = np.array([out.shape[0]]*out.shape[1],dtype = np.int64)
        out_len = torch.from_numpy(out_len).to(self.device)
        seq = batch['seq']
        seq_len = batch['seq_len'].view(-1)
        loss = net.ctc_loss(out,out_len,seq,seq_len)
        if self.config.TRAIN['label_smooth']:
            label_alternation = self.config.TRAIN['label_alternation'] 
            loss = net.ctc_loss_uncertain_label(out,out_len,seq,seq_len,label_alternation,self.config.TRAIN['label_smooth'])
        error = None
        if get_error:
            error = net.ctc_error(out,
                                  seq,
                                  seq_len,
                                  alphabet = 'N' + self.config.CTC['alphabeta'],
                                  beam_size = self.config.CTC['beam_size'],
                                  beam_cut_threshold = self.config.CTC['beam_cut_threshold'])
        return loss,error

def load_data(chunk_f,seq_f,seqlen_f,mmap_mode = "r",alphabeta = "ACGT",max_n = None):
    chunks = np.load(chunk_f,mmap_mode= mmap_mode)
    reference = np.load(seq_f,mmap_mode= mmap_mode)
    ref_len = np.load(seqlen_f,mmap_mode= mmap_mode)
    if len(chunks) > len(reference):
        print("There are more chunks (%d) than the sequences (%d), it will be cut to equal to the sequences."%(len(chunks),len(reference)))
        chunks = chunks[:len(reference)]
    if max_n is not None:
        chunks = chunks[:max_n]
        reference = reference[:max_n]
        ref_len = ref_len[:max_n]
    nan_filter = ~np.any(np.isnan(chunks),axis=1)
    chunks = chunks[nan_filter]
    reference = reference[nan_filter]
    ref_len = ref_len[nan_filter]
    ref_len = ref_len.astype(np.int64)
    if reference[0].dtype.kind in ['U','S']:
        alphabet_dict = {x:i+1 for i,x in enumerate(alphabeta)}
        dataset = Dataset(chunks,seq = reference,seq_len = ref_len,transform = transforms.Compose([NumIndex(alphabet_dict),ToTensor()]))
    else:
        dataset = Dataset(chunks,seq = reference,seq_len = ref_len,transform = transforms.Compose([ToTensor()]))
    return dataset

def main(args):
    if args.threads:
        torch.set_num_threads(args.threads)
    optimizers = {'Adam':torch.optim.Adam,
                  'AdamW':torch.optim.AdamW,
                  'SGD':torch.optim.SGD,
                  'RMSprop':torch.optim.RMSprop,
                  'Adagrad':torch.optim.Adagrad,
                  'Momentum':partial(torch.optim.SGD,momentum = 0.9)}
    class CTC_CONFIG(CONFIG):
        CTC = {"beam_size":5,
               "beam_cut_threshold":0.05,
               "alphabeta": "ACGTM",
               "mode":"rna"}
    class TRAIN_CONFIG(CTC_CONFIG):
        TRAIN = {"inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "grad_norm":2,
                 "keep_record":5,
                 "eval_size":100,
                 "optimizer":args.optimizer,
                 "embedding_pretrain_model":args.embedding,
                 "run_n_test":args.run_n_test,
                 "label_alternation":[5,1],#5 is the M label and 1 is the A label
                 "label_smooth":args.label_smooth}
    config = TRAIN_CONFIG()
    print("Construct and load the model.")
    model_f = args.model_folder
    if args.retrain:
        config_old = load_config(os.path.join(model_f,"config.toml"))
        config_old.TRAIN = config.TRAIN #Overwrite training config.
        config = config_old
    elif args.config:
        config = load_config(args.config)
    print("Read chunks and sequence.")
    dataset = load_data(args.chunks,args.seq,args.seq_len,mmap_mode = "r",alphabeta = config.CTC['alphabeta'])
    eval_size = config.TRAIN['eval_size']
    if args.eval_chunks:
        print("Read evaluation chunks and sequence.")
        eval_ds = load_data(args.eval_chunks,args.eval_seq,args.eval_seq_len,mmap_mode = "r",alphabeta = config.CTC['alphabeta'],max_n = eval_size*args.batch_size)
    else:
        dataset,eval_ds = torch.utils.data.random_split(dataset,[len(dataset) - eval_size, eval_size], generator=torch.Generator().manual_seed(42))
    loader = data.DataLoader(dataset,batch_size = config.TRAIN['batch_size'],shuffle = True, num_workers = 2)
    loader_eval = data.DataLoader(eval_ds,batch_size = config.TRAIN['batch_size'],shuffle = True, num_workers = 0)
    DEVICE = args.device
    loader = DeviceDataLoader(loader,device = DEVICE)
    loader_eval = DeviceDataLoader(loader_eval,device = DEVICE)
    print("Train dataset: %d batches; Evaluation dataset: %d batches"%(len(loader),len(loader_eval)))
    net = CRNN(config)
    if args.logging:
        writer = SummaryWriter(comment = "LR={LR}_MODEL={MODEL}_ReTrain={ReTrain}_ReOL={ReOL}".format(LR=args.lr,MODEL=os.path.basename(args.model_folder),ReTrain=args.retrain,ReOL=args.retrain_on_last))
        writer.add_text('chunks',args.chunks)
        writer.add_text('seq',args.seq)
    else:
        writer = None
    t = SupervisedTrainer(loader,net,config,eval_dataloader = loader_eval,logger = writer)
    if args.retrain:
        print("Load previous trained model.")
        t.load(model_f)
        if args.retrain_on_last:
            print("Freeze all layers except last %d fully-connected layers"%(args.retrain_on_last))
            assert args.retrain_on_last <= config.FNN['N_Layer'], "Argument retrain_on_last %d is larger than the number of FNN layers %d."%(args.retrain_on_last,config.FNN['N_Layer'])
            n_skip = sum([2 if layer['dropout'] is None else 3 for layer in config.FNN['Layers'][:-args.retrain_on_last] ])
            for layer in t.net.net[:-n_skip]:
                #multiply by 2 here because we have weight and bias.
                for param in layer.parameters():
                    param.requires_grad = False
        if args.reinitialize_methylation:
            print("Reinitialize the methyaltion projection weight.")
            projection = t.net.net[-2]
            stdv = 1./math.sqrt(projection.weight.size(1))
            with torch.no_grad():
                projection.weight[-1] = (torch.rand(projection.weight[-1].shape)-0.5)*stdv
                if projection.bias is not None:
                    projection.bias[-1] = (torch.rand(1)-0.5)*stdv
    elif args.embedding:
        print("Import the embedding model from %s"%(args.embedding))
        t.load(args.embedding)
        for layer in t.net.embedding_layers:
            for param in layer.parameters():
                param.requires_grad = False
    lr = args.lr
    epoches = args.epoches
    optim = optimizers[config.TRAIN['optimizer']](net.parameters(),lr = lr)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optim, 'min', patience = 10,factor = 0.3,verbose = True)
    # scheduler = None
    COUNT_CYCLE = args.report
    valid_check_cycle = args.valid_cycle
    print("Begin training the model.")
    t.train(epoches,optim,COUNT_CYCLE,valid_check_cycle,model_f,scheduler = scheduler)
    
def add_arguments(parser):
    parser.add_argument('-i', '--chunks', required = True,
                        help = "The .npy file contain chunks.")
    parser.add_argument('-o', '--model_folder', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument('--seq', required = True,
                        help="The .npy file contain the sequence.")
    parser.add_argument('--seq_len', required = True,
                        help="The .npy file contain the sueqnece length.")
    parser.add_argument('--eval_chunks',default = None,
                        help="The .npy file contain the chunks for evaluation.")
    parser.add_argument('--eval_seq',default = None,
                        help="The .npy file contain the sequence for evaluation.")
    parser.add_argument('--eval_seq_len',default = None,
                        help="The .npy file contain the sequence length for evaluation.")
    parser.add_argument('--embedding', default = None,
                        help="The folder contains the embedding model, if retrain is enable this will have no effect.")
    parser.add_argument('--device', default = 'cuda',
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--lr', default = 4e-3, type = float,
                        help="Initial learning rate.")
    parser.add_argument('--batch_size', default = 200, type = int,
                        help="Training batch size.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument('--report', default = 20, type = int,
                        help = "The interval of training rounds to report.")
    parser.add_argument('--valid_cycle',default = 1000, type = int,
                        help = "The number of cycles to validate the model.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--config', default = None,
                        help = "Training configuration.")
    parser.add_argument('--optimizer', default = "RMSprop",
                        help = "Optimizer to use, can be Adam, AdamW, SGD and RMSprop,\
                            default is RMSprop")
    parser.add_argument('--threads', type = int, default = None,
                        help = "Number of threads used by Pytorch")
    parser.add_argument('--run_n_test', type = int, default = None,
                        help = "Number of batches to run for test if not None.")
    parser.add_argument('--retrain_on_last',type = int, default = None,
                        help = "Number of FNN layers to retrain on, the rest of the layers will be freezed.")
    parser.add_argument('--reinitialize_methylation', dest = "reinitialize_methylation", action = "store_true",
                        help = "Reinitialize the methylation base projection in last FNN layer.")
    parser.add_argument('--nolog', action="store_false", dest="logging",
                        help = "Disable logging.")
    parser.add_argument('--label_smooth', type = float, default = 0.0)

def post_args(args):
    if args.logging:
        from torch.utils.tensorboard import SummaryWriter
    if args.retrain and args.embedding:
        args.embedding = None
        print("Embedding is being overrided by --load argument.")
    os.makedirs(args.model_folder,exist_ok=True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training model with tfrecord file')
    args = parser.parse_args(sys.argv[1:])
    main(args)



#This script is used to prepare training dataset for methylation basecaller by extracting raw signal chunks with m6a segments
import re
import os
import h5py
import pysam
import editdistance
from Bio import motifs
from Bio.Seq import Seq 
from sklearn.mixture import GaussianMixture
from xron.utils.seq_op import norm_by_noisiest_section
import numpy as np
import pandas as pd
from tqdm import tqdm

motif_reg = {"A":"A","C":"C","G":"G","T":"T","U":"U","R":"[AG]","Y":"[CT]","M":"[AC]","K":"[GT]","S":"[GC]","W":"[AT]","B":"[CGT]","D":"[AGT]","H":"[ACT]","V":"[ACG]","N":"[ACGT]"}
def get_reg(motif):
    reg = ""
    for i in motif:
        reg += motif_reg[i]
    return reg

def rc(seq):
    seq = seq.upper()
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}
    return ''.join([complement[base] for base in reversed(seq)])

def parse_qr_string(qr_string):
    q,r = qr_string.split('|')
    q = np.asarray([int(float(x)) for x in q.split(':')])
    r = np.asarray([int(float(x)) for x in r.split(':')])
    # r = r-r[0] #Make the first index 0
    return q,r

def exclude_unmached_read(df):
    """Exclude reads whose basecall sequence is not the same as the query sequence
    This is usually due to there is repeating read name id.
    """
    dist = np.asarray([editdistance.eval(rc(x),y) if z else editdistance.eval(x,y) for x,y,z in tqdm(zip(df['query_seq'].values,df['basecall_seq'].values,df['reverse_align'].values))])
    gmm = GaussianMixture(n_components=2)
    gmm.fit(dist.reshape(-1,1))
    cluster_means = gmm.means_
    threshold = np.max([np.percentile(cluster_means,95),np.mean(cluster_means)])
    mask = dist < threshold
    df = df[mask]
    return df

def get_ref_idx(q,r,q_idx):
    return r[np.where(q>=q_idx)[0][0]]

#Set a repository name in the system environment variable NAME
CONFIG = "HEK293T"
BRIDGE_FOLDER = os.environ['SCRATCH']+'/'
if CONFIG == "Yearst":
    repo = "xron_crosslink2"
    names = ["ko1","ko2","ko3","wt1","wt2","wt3"]
    meths = [False,False,False,True,True,True]
    reference = f"{BRIDGE_FOLDER}/ime4_Yearst/Yearst/Yeast_sk1.fasta"
    reference = pysam.FastaFile(reference)
    segment_len = 4000
    window_half = 4
    xron_basecall_config = {"stride":11,
                            "move_forward":False,}
    motif_string = "DRACD"
    motif_A_idx = 2
    BASE_F = f"{BRIDGE_FOLDER}/Xron_Project/Benchmark/Yearst"

if CONFIG == "HEK293T":
    repo = "xron_crosslink2_YeastFinetune_4000L"
    # names = ["wt1-balance_chr1",
    #          "wt1-balance_chr2",
    #          "wt1-balance_chr3",
    #          "wt1-balance_chr4",
    #          "wt1-balance_chr5",
    #          "wt1-balance_chr6",
    #          "wt1-balance_chr7",
    #          "wt1-balance_chr8",
    #          "wt1-balance_chr9",
    #          "wt1-balance_chr10",
    #          "wt1-balance_chr11",
    #          "wt1-balance_chr12",
    #          "wt1-balance_chr13",
    #          "wt1-balance_chr14",
    #          "wt1-balance_chr15",
    #          "wt1-balance_chr16",
    #          "wt1-balance_chr17",
    #          "wt1-balance_chr18",
    #          "wt1-balance_chr19",
    #          "wt1-balance_chr20",
    #          "wt1-balance_chr21",
    #          "wt1-balance_chr22",
    #          "wt1-balance_chrX",
    #          "wt1-balance_chrM",]
    # names = ["wt1-balance_chr3",
    #          "wt1-balance_chr4",
    #          "wt1-balance_chr5",
    #          "wt1-balance_chr6",
    #          "wt1-balance_chr7",
    #          "wt1-balance_chr8",
    #          "wt1-balance_chr9",
    #          "wt1-balance_chr10",
    #          "wt1-balance_chr11",
    #          "wt1-balance_chr12",
    #          "wt1-balance_chr13"]
    # names = ["wt1-balance_chr14",
    #          "wt1-balance_chr15",
    #          "wt1-balance_chr16",
    #          "wt1-balance_chr17",
    #          "wt1-balance_chr18",
    #          "wt1-balance_chr19",
    #          "wt1-balance_chr20",
    #          "wt1-balance_chr21",
    #          "wt1-balance_chr22",
    #          "wt1-balance_chrX",
    #          "wt1-balance_chrM"]
    names = [os.environ['NAME']]
    #TODO change the name
    meths = [None]*len(names)
    BASE_F = f"{BRIDGE_FOLDER}/Xron_Project/Benchmark/HEK293T/xron_crosslink2_finetune/"
    sites_f = f"{BRIDGE_FOLDER}/Xron_Project/m6A_site_m6Anet_DRACH_HEK293T.csv"
    sites = pd.read_csv(sites_f)
    reference = f"{BRIDGE_FOLDER}/NA12878_RNA_IVT/GRCh38_transcript_ensembel_sgnex/Homo_sapiens.GRCh38.cdna.ncrna.fa"
    reference = pysam.FastaFile(reference)
    segment_len = 4000
    window_half = 4
    xron_basecall_config = {"stride":11,
                            "move_forward":False,}
    motif_string = "DRACD"
    motif_A_idx = 2


for name,meth in zip(names,meths):
    input_site_summary = f"{BASE_F}/readIDs_{repo}_{name}.csv"
    out_f = f"{BASE_F}/datasets/{repo}_{name}_{motif_string}/"
    os.makedirs(out_f,exist_ok=True)
    print(f"Loading summary {name}...")
    config = xron_basecall_config
    df = pd.read_csv(input_site_summary)
    df = df.dropna()
    print("Excluding unmatched reads...")
    df = exclude_unmached_read(df)
    print("Loading reference...")
    chunks = []
    seqs = []
    chunk_lens = []
    seq_lens = []
    motif_segments = []
    summary = {"succeed":0,"Broken basecall entry":0, "Sampling sequence out of alignment bounds":0,"Motif is missing":0}

    if type(motif_string) is str:
        motif = re.compile(get_reg(motif_string))
    print("Extracting segments...")
    with tqdm(total = len(df)) as pbar:
        for f in set(df['fast5s'].values):
            ids = df[df['fast5s'] == f]['ids'].values
            cols = df[df['fast5s'] == f].index
            with h5py.File(f.replace("/ocean/projects/hmcmutc/haotiant/",f"{BRIDGE_FOLDER}"),'r') as f5:
                for id,idx in zip(ids,cols):
                    q,r = parse_qr_string(df['qr_index'][idx])
                    refpos = df['refpos'][idx]-1
                    read_h = f5[f"read_{id}"]
                    result_h = read_h['Analyses/Basecall_1D_000/BaseCalled_template/']
                    signal = np.asarray(read_h['Raw/Signal']) 
                    original_signal = signal.astype(np.float32)
                    signal,med,mad = norm_by_noisiest_section(signal)
                    signal = signal.astype(np.float16)[::-1]#Reverse the signal because RNA is 5'->3'
                    seq = str(np.asarray(result_h['Fastq']).astype(str)).split('\n')[1]
                    move = np.asarray(result_h['Move'])
                    if len(move) == 0:
                        summary["Broken basecall entry"] += 1
                        pbar.update()
                        continue
                    if config['move_forward']:
                        move = move[::-1]

                    #build the signal index of the sequence with 11 increase step from move matrix
                    signal_idx = np.where(move == 1)[0]*config['stride']

                    #Sample a signal segment of 4000 length around the m6A site
                    upstream = np.random.randint(0,segment_len - 200)
                    q_seq = df['query_seq'][idx]
                    q_pos = df['seq_pos'][idx]
                    reverse = df['reverse_align'][idx]
                    bc_pos = q_pos if not reverse else len(q_seq) - q_pos
                    try:
                        start = max(signal_idx[bc_pos] - upstream,0)
                    except IndexError:
                        summary["Broken basecall entry"] += 1
                        pbar.update()
                        continue
                    bc_start,bc_end = np.where(signal_idx>=start)[0][0],np.where(signal_idx<=start+segment_len)[0][-1]
                    q_start = bc_start if not reverse else len(q_seq) - bc_end
                    q_end = bc_end if not reverse else len(q_seq) - bc_start
                    if q_start - q[0] < -window_half or q_end - q[-1] > window_half:
                        summary["Sampling sequence out of alignment bounds"] += 1
                        pbar.update()
                        continue
                    q_start = np.max([q_start,q[0]])
                    q_end = np.min([q_end,q[-1]])
                    ref_start = min(get_ref_idx(q,r,q_start),refpos-1)
                    ref_end = max(get_ref_idx(q,r,q_end),refpos+1)
                    if abs(ref_end - ref_start - (q_end - q_start)) > 10:
                        summary["Sampling sequence out of alignment bounds"] += 1
                        pbar.update()
                        continue
                    try:
                        ref_seq = reference.fetch(df['contig'][idx],ref_start,ref_end+1)
                    except ValueError as e:
                        print(e)
                        summary["Sampling sequence out of alignment bounds"] += 1
                        pbar.update()
                        continue
                    refpos = refpos - ref_start #Relative position to the start of the segment
                    if reverse:
                        ref_seq = rc(ref_seq)
                        refpos = len(ref_seq) - refpos
                    if ref_seq[refpos] != 'A':
                        motif_hit = re.search(motif,ref_seq[refpos-window_half:refpos+window_half+1])
                        if not motif_hit:
                            summary["Motif is missing"] += 1
                            pbar.update()
                            continue
                        else:
                            refpos += motif_hit.start() - window_half + motif_A_idx #Fixing some minor shiftings
                    motif_segments.append(ref_seq[refpos-window_half:refpos+window_half+1])
                    if meth is not None:
                        if meth:
                            #replace A with M at refpos to indicate the methylation
                            ref_seq = ref_seq[:refpos] + 'M' + ref_seq[refpos+1:]
                    else:
                        entry = sites[(sites['chr'] == df['contig'][idx]) & (sites['genomic_position'] == df['refpos'][idx])]
                        if len(entry) == 0:
                            summary["Motif is missing"] += 1
                            continue
                        if entry['modification_status'].values[0] == 1:
                            ref_seq = ref_seq[:refpos] + 'M' + ref_seq[refpos+1:]
                    seq_lens.append(len(ref_seq))
                    seqs.append(ref_seq)
                    curr_seg = signal[start:start+segment_len]
                    chunk_lens.append(len(curr_seg))
                    if len(curr_seg) < segment_len:
                        curr_seg = np.pad(curr_seg,(0,segment_len-len(curr_seg)),'constant',constant_values=0)
                    chunks.append(curr_seg)
                    summary["succeed"] += 1
                    pbar.set_postfix({"succeed":summary["succeed"],
                                    "Broken basecall entry":summary["Broken basecall entry"],
                                    "Sampling sequence out of alignment bounds":summary["Sampling sequence out of alignment bounds"],
                                    "Motif is missing":summary["Motif is missing"]})
                    pbar.update()
    print("Saving...")
    np.save(out_f + "seqs.npy",seqs)
    np.save(out_f + "chunk_lens.npy",chunk_lens)
    np.save(out_f + "seq_lens.npy",seq_lens)
    np.save(out_f + "chunks.npy",np.asarray(chunks))
    motif_segments = [Seq(x) for x in motif_segments if len(x) == (window_half*2+1) ]
    motif = motifs.create(motif_segments)
    print("Done!")



"""
Created on Tue Apr 19 16:19:13 2022

@author: Haotian Teng
"""
import os
import sys
import time
import toml
import torch
import argparse
import numpy as np
from tqdm import tqdm
from typing import List, Dict
from functools import partial
from xron.utils.seq_op import length_mask
from xron.nrhmm.hmm import RHMM,GaussianEmissions
from xron.nrhmm.hmm_input import Kmer2Transition, Kmer_Dataset, Normalizer
from xron.nrhmm.hmm_relabel import get_effective_kmers
from torchvision import transforms
from torch.utils.data.dataloader import DataLoader
from xron.xron_train_base import Trainer,DeviceDataLoader

class RHMM_Trainer(Trainer):
    def __init__(self,
                 train_dataloader:DataLoader,
                 hmm:RHMM,
                 config:object):
        super().__init__(train_dataloader,
                         nets = {"hmm":hmm},
                         config = config,
                         device = config.TRAIN["device"]
                         )
        self.hmm = hmm
        
    def train(self,epoches,optimizer,save_cycle,save_folder):
        self.save_folder = save_folder
        self._save_config()
        for epoch_i in range(epoches):
            for i_batch, batch in enumerate(self.train_ds):
                start = time.time()
                loss = self.train_step(batch)
                if torch.isnan(loss):
                    print("NaN loss detected, skip this training step.")
                    continue
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                self.hmm.emission.clamp_cov()
                self.losses.append(loss.item())
                self.global_step +=1
                if (i_batch+1)%save_cycle==0:
                    time_elapsed = time.time() - start
                    mse_loss = self.eval_step(batch)
                    self.save()
                    print("Epoch %d: Batch %d, loss %f, rc loss %.2f, time elapsed per batch %.2f"%(epoch_i, i_batch, loss, mse_loss, time_elapsed))
                    self.save_loss()

    def train_step(self,batch):
        net = self.hmm
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        if torch.sum(torch.isnan(signal_batch)):
            print("Found NaN input signal.")
            return None
        log_prob = net.forward(signal_batch, duration_batch, transition_batch)
        loss = -torch.logsumexp(log_prob, dim = 1).mean()
        return loss
    
    def eval_step(self,batch):
        net = self.hmm
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        if torch.sum(torch.isnan(signal_batch)):
            print("Found NaN input signal.")
            return None
        path,logit = net.viterbi_decode(signal_batch, duration_batch, transition_batch)
        rc_signal = np.asarray([[net.emission.means[x].item() for x in p] for p in path.cpu().numpy()])
        diff = (rc_signal - signal_batch[:,:,0].cpu().numpy())**2
        loss = np.mean(diff[length_mask(duration_batch.cpu().numpy(),signal_batch.shape[1])])
        return loss
    
    def train_EM(self,epoches,save_cycle,save_folder):
        self.save_folder = save_folder
        self._save_config()
        for epoch_i in range(epoches):
            for i_batch, batch in enumerate(self.train_ds):
                start = time.time()
                renew = ((i_batch)%save_cycle==0)
                self.train_step_EM(batch,update_parameter = renew)
                self.global_step +=1
                if renew:
                    time_elapsed = time.time() - start
                    with torch.no_grad():
                        loss = self.train_step(batch).detach()
                        if loss is None:
                            print("NaN loss detected, skip this training step.")
                            continue
                    mse_loss = self.eval_step(batch)
                    self.losses.append(loss.item())
                    self.save()
                    print("Epoch %d: Batch %d, loss %f, MSE %.2f, time elapsed per batch %.2f"%(epoch_i, i_batch, loss, mse_loss, time_elapsed))
                    self.save_loss()
                if (i_batch+1)%self.config.TRAIN['cov_update']==0:
                    self.hmm.emission.update_cov(exploration = self.config.TRAIN["exploration"])
    
    def train_step_EM(self,batch,update_parameter):
        hmm = self.hmm
        signal_batch = batch['signal']
        duration_batch = batch['duration']
        transition_batch = batch['labels']
        if torch.sum(torch.isnan(signal_batch)):
            print("Found NaN input signal.")
            return None
        with torch.no_grad():
            gamma = hmm.expectation(signal_batch, duration_batch, transition_batch)
            hmm.maximization(signal_batch,
                             gamma,
                             update = update_parameter,
                             lr=self.config.TRAIN["learning_rate"],
                             weight_decay = self.config.TRAIN["weight_decay"],
                             momentum = self.config.TRAIN["momentum"])
        return gamma.detach()

def trainable_kmer(idx2kmer:List,must_contain:str,exclude_bases:str)->np.array:
    """
    Generate the trainble mask given the trainable bases 

    Parameters
    ----------
    idx2kmer : List
        The index to kmer map.
    must_contain : str, optional
        Bases that must exist in the kmer to make the kmer trainable.
    exlcude_bases : str, optional
        Bases that should NOT exist in the kmer to make the kmer trainable.

    Returns
    -------
    np.array
        A length N array contains the boolean variable that mark the trainble kmers, N is the number of kmer.

    """
    trainable = np.ones(len(idx2kmer),dtype = bool)
    for i,kmer in enumerate(idx2kmer):
        for c in kmer:
            if c in exclude_bases:
                trainable[i] = False
                break
        for b in must_contain:
            if b not in kmer:
                trainable[i] = False
                break
    return trainable

def initialize_modified_kmers(idx2kmer:List, 
                              emission_means:np.array, 
                              shift:float = 0.0,
                              modified_bases:Dict = {'A':'M'})->np.array:
    """
    Generate the initialization of the modified kmers.

    Parameters
    ----------
    idx2kmer : List
        The index to kmer look-up list.
    emission_means : np.array
        The current mean of the kmers.
    shift : float, optional
        The prior shift for the modified kmer. The default is -0.3.
    modified_bases : Dict, optional
        The dictionary of alphabeta of all modified bases, default is {'M':'A'}.

    """
    modified_means = np.copy(emission_means)
    for i,kmer in enumerate(idx2kmer):
        if any([b in kmer for b in modified_bases.values()]):
            orig_kmer = kmer
            for key,val in modified_bases.items():
                orig_kmer = orig_kmer.replace(val,key)
            modified_means[i] = emission_means[idx2kmer.index(orig_kmer)]+shift
    return modified_means

def train(args):
    optimizers = {'Adam':torch.optim.Adam,
                  'AdamW':torch.optim.AdamW,
                  'SGD':torch.optim.SGD,
                  'RMSprop':torch.optim.RMSprop,
                  'Adagrad':torch.optim.Adagrad,
                  'Momentum':partial(torch.optim.SGD,momentum = 0.9)}
    config = toml.load(os.path.join(args.input,"config.toml"))
    class TRAIN_CONFIG(object):
        DATA = config
        TRAIN = {"keep_record":5,
                 "grad_norm":2,
                 "inital_learning_rate":args.lr,
                 "batch_size":args.batch_size,
                 "optimizer":optimizers[args.optimizer],
                 "device":args.device,
                 "learning_rate":args.lr,
                 "weight_decay":args.weight_decay,
                 "momentum":args.momentum,
                 "exploration":args.exploration,
                 "cov_update":args.reestimate_covariance,
                 "transition_type":args.transition_format}
    train_config = TRAIN_CONFIG()
    MODIFIED_BASES = {"A":"M"}
    if args.trainable_bases is None:
        args.trainable_bases = "!"
    train_config.TRAIN["trainable_bases"] = args.trainable_bases
    
    ### Read dataset
    print("Loading dataset.")
    if 'base_prior' not in config.keys():
        config['base_prior'] = None
    if args.methylation_proportion is not None:
        config['base_prior'] = {"M":args.methylation_proportion,
                                "A":1-args.methylation_proportion}
        #TODO: this has to be changed latter to adapt to multi base-modifications.
    chunks = np.load(os.path.join(args.input,"chunks.npy"))
    nan_filter = ~np.any(np.isnan(chunks),axis=1)
    durations = np.load(os.path.join(args.input,"durations.npy"))
    kmers = np.load(os.path.join(args.input,"kmers.npy"))
    chunks,durations,kmers = chunks[nan_filter],durations[nan_filter],kmers[nan_filter]
    print("Dataset loaded, begin transfer.")
    k2t = Kmer2Transition(config['alphabeta'],
                          config['k'],
                          config['chunk_len'],
                          config['kmer2idx_dict'],
                          config['idx2kmer'],
                          base_alternation = MODIFIED_BASES, 
                          base_prior = config['base_prior'],
                          kmer_replacement = args.kmer_replacement,
                          out_format = "compact" if train_config.TRAIN['transition_type'] == "compact" else "sparse")
    dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
    loader = DataLoader(dataset,batch_size = args.batch_size, shuffle = True,num_workers=5)
    loader = DeviceDataLoader(loader,device = args.device)
    
    ## Initialize the model
    print("Construct and initialize the model.")
    n_states = len(config['alphabeta'])**config['k']
    init_means = np.random.rand(n_states)-0.5
    init_covs = np.asarray([0.5]*n_states)
    
    ### Initialize parameter
    if args.initialize:
        N,L = chunks.shape
        duration_mask = np.repeat(np.arange(L)[None,:],N,axis = 0)<durations[:,None]
        if not args.retrain:
            for i in tqdm(np.arange(n_states),desc = "Initialize parameter from the given chunks."):
                selected_sig = chunks[np.logical_and(duration_mask,kmers == i)]
                if len(selected_sig) > 0:
                    init_means[i] = np.mean(selected_sig)
            np.save(os.path.join(args.model_folder,"initial_means0.npy"),init_means)
            print("Rescale the signal using the initialized parameters.")
        
    ### Building model
    trainable = trainable_kmer(train_config.DATA['idx2kmer'],*train_config.TRAIN["trainable_bases"].split('!'))
    emission = GaussianEmissions(init_means[:,None], init_covs[:,None],trainable = trainable)
    hmm = RHMM(emission,
               normalize_transition=False,
               device = args.device,
               transition_operation=train_config.TRAIN['transition_type'],
               index_mapping = None if train_config.TRAIN['transition_type'] != "compact" else k2t.idx_map)
    
    ## Train
    trainer = RHMM_Trainer(loader,hmm,config = train_config)
    if args.retrain:
        print("Load pretrained model.")
        trainer.load(args.model_folder)
        init_means = trainer.hmm.emission.means.cpu().detach().numpy()
        init_covs = trainer.hmm.emission.cov.cpu().detach().numpy()
    lr = args.lr
    epoches = args.epoches
    optim = train_config.TRAIN['optimizer'](hmm.parameters(),lr = lr)
    
    ### Preprocess before training
    if args.reinitialize:
        print("Initialize the modified kmers according to the current kmer model.")
        modified_means = initialize_modified_kmers(config['idx2kmer'],
                                                   trainer.hmm.emission.means.cpu().detach().numpy(),
                                                   modified_bases=MODIFIED_BASES)
        trainer.hmm.emission.reinitialize_means(torch.from_numpy(modified_means).to(args.device))
    
    ### Renormalize the signal
    if args.effective == "!":
        effective_kmers = None
    else:
        effective_kmers = get_effective_kmers(args.effective, 
                                              train_config.DATA['idx2kmer'])
    norm = Normalizer(use_dwell = True,
                      no_scale = True,
                      effective_kmers=effective_kmers)
    if args.pretrain > 0:
        renorm_chunks = chunks
        for j in tqdm(np.arange(args.pretrain),desc = "Renorm the signal"):
            rc_signal = np.asarray([[init_means[x].item() for x in kmers[i]] for i in tqdm(np.arange(chunks.shape[0]),desc = "Reconstruct signal")])
            print("Renormalization.")
            renorm_chunks = norm(renorm_chunks,rc_signal, durations,kmers)
            print("Save the renormalized chunks.")
            np.save(os.path.join(args.input,"chunks_renorm_%d.npy"%(j)),renorm_chunks)
            chunks = renorm_chunks
    renorm_dataset = Kmer_Dataset(chunks, durations, kmers,transform=transforms.Compose([k2t]))
    loader = DataLoader(renorm_dataset,batch_size = args.batch_size, shuffle = True,num_workers=args.threads)
    loader = DeviceDataLoader(loader,device = args.device)
    trainer.reload_data(loader)
    
    print("Begin training the model.")
    trainer.train_EM(epoches,args.report,args.model_folder)
    # trainer.train(epoches,optim,args.report,args.model_folder)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training RHMM model')
    parser.add_argument("-i","--input", type = str, required = True,
                        help = "Data folder contains the chunk, kmer sequence.")
    parser.add_argument('-o', '--model_folder', required = True,
                        help = "The folder to save folder at.")
    parser.add_argument("-b","--batch_size", type = int, default = 30,
                        help = "The batch size to train.")
    parser.add_argument("--lr",type = float, default = 1.0,
                        help = "The initial training learning rate.")
    parser.add_argument("--weight_decay",type = float, default = 0.0,
                        help = "The weight decay of the parameters.")
    parser.add_argument("--momentum",type = float, default = 0.9,
                        help = "The momentum value.")
    parser.add_argument("--exploration",type = float, default = 1.5,
                        help = "The factor of setting covariance to certainty, bigger number make the model explore more space of mean but may decrease the convergence speed.")
    parser.add_argument("--transition_format",type = str, default = "sparse",
                        help = "The format that being used to calculate the transition step, can be sparse, dense and compact, default is sparse.")
    parser.add_argument("--report",type = int, default = 10,
                        help = "Report the loss and save the model every report cycle.")
    parser.add_argument("--reestimate_covariance", type = int, default = 20,
                        help = "Reestimate the covariance every N batches.")
    parser.add_argument("--certain_methylation",action = "store_false", 
                        dest = "kmer_replacement", 
                        help = "If we are sure about the methylation state.")
    parser.add_argument("--optimizer",type = str, default = "Adam",
                        help = "The optimizer used to train the model.")
    parser.add_argument('--epoches', default = 10, type = int,
                        help = "The number of epoches to train.")
    parser.add_argument("--device",type = str, default = "cuda",
                        help = "The device used to train the model.")
    parser.add_argument('--load', dest='retrain', action='store_true',
                        help='Load existed model.')
    parser.add_argument('--trainable_bases', type = str, default = None,
                        help="A magic string AB!CD or AB that gives the trainable bases and NOT trainalbe bases, for example MC means trains on kmer that must contains M and C, A!M means trains on kmer that must contains A but not contains M.")
    parser.add_argument('--methylation_proportion', type = float, default = None,
                        help="The expecting methylation proportion.")
    parser.add_argument('--pretrain', type = int, default = 0,
                        help="The rounds to renormalize signal and pretrain.")
    parser.add_argument("-e","--effective",type = str, default = "!",
                        help = "A magic string gives the kmers that take into \
                            account when doing normalization, for example, \
                            A!M means kmers that must have A and must not \
                            have M is taken into account.")
    parser.add_argument('--initialize_modified_kmers', dest = 'reinitialize',
                        action = "store_true", help = "Reinitialize the modifed kmers.")
    parser.add_argument('--no_initialize', dest = "initialize",
                        action = "store_false", help = "Turn off the parameters initialization.")
    parser.add_argument('--threads',type = int, default = 0,
                        help = "The number of threads used to load the data.")
    args = parser.parse_args(sys.argv[1:])
    os.makedirs(args.model_folder,exist_ok=True)
    train(args)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 15 00:04:58 2021

@author: haotian teng
"""
import os
import sys
import umap
import torch
import argparse
import numpy as np
import torch.utils.data as data
from itertools import islice
from typing import List,Union
from torchvision import transforms
from matplotlib import pyplot as plt
from xron.xron_input import Dataset, ToTensor
from xron.xron_train_base import Trainer, load_config, DeviceDataLoader
from xron.xron_model import REVCNN,DECODER_CONFIG,CRNN,MM
from torch.distributions.one_hot_categorical import OneHotCategorical as OHC

class Evaluator(Trainer):
    def __init__(self, 
                 encoder:CRNN, 
                 decoders:List[Union[REVCNN,MM]],
                 config:DECODER_CONFIG,
                 device:str = None):
        device = config.EVALUATION['device']
        super().__init__(train_dataloader=None,
                       nets = {"encoder":encoder,
                               "decoder":decoders[0],
                               "mm":decoders[1]},
                       config = config,
                       device = device,
                       eval_dataloader = None)
        self.encoder = encoder
        self.decoder_revcnn, self.decoder_mm = decoders
        
    def eval_once(self,batch:np.ndarray):
        encoder = self.encoder
        d1 = self.decoder_revcnn
        d2 = self.decoder_mm
        signal = batch['signal']
        logprob = encoder.forward(signal) #[L,N,C]
        prob = torch.exp(logprob)
        m = OHC(prob)
        sampling = m.sample().permute([1,2,0]) #[L,N,C]->[N,C,L]
        rc_signal = d1.forward(sampling).permute([0,2,1]) #[N,L,C] -> [N,C,L]
        if d2:
            rc_signal += d2.forward(sampling,device = self.device).permute([0,2,1])
        predictions = encoder.ctc_decode(logprob,
                       alphabet = 'N' + self.config.CTC['alphabeta'],
                       beam_size = self.config.CTC['beam_size'],
                       beam_cut_threshold = self.config.CTC['beam_cut_threshold'])
        return rc_signal,prob, predictions,sampling

class VQ_Evaluator(Trainer):
    def __init__(self, 
                 encoder:CRNN, 
                 decoders:List[Union[REVCNN,MM]],
                 config:DECODER_CONFIG,
                 device:str = None):
        device = config.EVALUATION['device']
        super().__init__(train_dataloader=None,
                       nets = {"encoder":encoder,
                               "decoder":decoders[0],
                               "mm":decoders[1]},
                       config = config,
                       device = device,
                       eval_dataloader = None)
        self.encoder = encoder
        self.decoder_revcnn, self.decoder_mm = decoders
        self.umap_transformer = umap.UMAP()
        
    def eval_once(self,batch:np.ndarray,umap_visualize = True):
        encoder = self.encoder
        d1 = self.decoder_revcnn
        d2 = self.decoder_mm
        # embedding = d2.level_embedding
        signal = batch['signal']
        embed = encoder.forward_wo_fnn(signal) #[L,N,C]
        if umap_visualize:
            u = self.umap_embedding(embed)
        rc_signal = d1.forward(embed).permute([0,2,1]) #[N,L,C] -> [N,C,L]
        return rc_signal,u
    
    def umap_embedding(self,embedding:np.ndarray):
        u = self.umap_transformer.fit_transform(embedding.view(-1,embedding.shape[-1]).detach().cpu().numpy())
        return u
    

def cmd_args():
    parser = argparse.ArgumentParser(
        description='Training model with tfrecord file')
    parser.add_argument('-i', '--chunks', required = True,
                        help = "The .npy file contain chunks.")
    parser.add_argument('--model_folder', required = True,
                        help = "The folder contains the trained model.")
    parser.add_argument('--seq', default = None,
                        help="The .npy file contain the sequence.")
    parser.add_argument('--seq_len', default = None,
                        help="The .npy file contain the sueqnece length.")
    parser.add_argument('--device', default = 'cuda',
                        help="The device used for training, can be cpu or cuda.")
    parser.add_argument('--repeat', type = int, default = 5,
                        help="The repeat used to test.")
    parser.add_argument('--method',default = "VQ",
                        help="The embedding method used to train, can be VQ or MM")
    args = parser.parse_args(sys.argv[1:])
    return args

if __name__ == "__main__":
    args = cmd_args()
    
    #Load model
    print("Load model.")
    config = load_config(os.path.join(args.model_folder,'config.toml'))
    stride = config.CNN['Layers'][-1]['stride']
    config.EVALUATION = {"batch_size":100,
                         "device":args.device}
    encoder = CRNN(config)
    revcnn = REVCNN(config) if 'CNN_DECODER' in config.__dict__.keys() else None
    mm = MM(config) if 'PORE_MODEL' in config.__dict__.keys() else None
    if args.method == "VQ":
        e = VQ_Evaluator(encoder,[revcnn,mm],config,device = args.device)
    elif args.method == "MM":
        e = Evaluator(encoder,[revcnn,mm],config,device = args.device)
    e.load(args.model_folder)
    
    #Load data
    print("Load data.")
    chunks = np.load(args.chunks)
    if args.seq:
        reference = np.load(args.seq)
    else:
        reference = None
    if args.seq_len:
        ref_len = np.load(args.seq_len)
    else:
        ref_len = None
    dataset = Dataset(chunks,seq = reference,seq_len = ref_len,transform = transforms.Compose([ToTensor()]))
    loader = data.DataLoader(dataset,batch_size = config.EVALUATION["batch_size"],shuffle = True, num_workers = 4)
    DEVICE = args.device
    loader = DeviceDataLoader(loader,device = DEVICE)
    
    #Evaluation
    batch = next(islice(loader,2,None))
    rc_signal, umap_vis = e.eval_once(batch)
    rc_signal = rc_signal.detach().cpu().numpy()
    norm_signal = (rc_signal - np.mean(rc_signal,axis = 2))/np.std(rc_signal,axis = 2)
    
    #Plot
    for r in np.arange(args.repeat):
        idx = np.random.randint(low = 0, high = config.EVALUATION['batch_size']-1)
        fig,axs = plt.subplots(nrows = 2,figsize = (20,30),gridspec_kw={'height_ratios': [1, 2]})
        start_idx =0
        last_idx = 800
        axs[0].plot(norm_signal[idx,0,start_idx:last_idx],label = "Reconstruction")
        axs[0].plot(batch['signal'].cpu()[idx,0,start_idx:last_idx],label = "Original signal")
        axs[0].legend()
        axs[1].scatter(umap_vis[:,0],umap_vis[:,1])
        fig.savefig(os.path.join(args.model_folder,'reconstruction_%d.png'%(r)))
        


"""
Created on Fri Aug 13 16:03:48 2021

@author: Haotian Teng
"""
import os,sys
import numpy as np
from matplotlib import pyplot as plt
import argparse
def map_base(string):
    base_dict = {'A':1,'C':2,'G':3,'T':4}
    return np.asarray([base_dict[b] for b in string])


#bonito_ref = "/home/heavens/twilight_data1/S10_DNA/20170322_c4_watermanag_S10/bonito_output/ctc_data/references.npy"
#bonito_len = "/home/heavens/twilight_data1/S10_DNA/20170322_c4_watermanag_S10/bonito_output/ctc_data/reference_lengths.npy"
#ref = np.load(bonito_ref)
#ref_len = np.load(bonito_len)
#plt.hist(ref_len)
def main(args):
    print("Read the chunks...") 
    chunk_f = os.path.join(args.input,'chunks.npy')
    seq_f = os.path.join(args.input,'seqs.npy')
    seq_len_f = os.path.join(args.input,'seq_lens.npy')
    seq = np.load(seq_f)
    seq_len = np.load(seq_len_f)
    chunk = np.load(chunk_f)
    print("Process...")
    plt.figure()
    plt.hist(seq_len,bins = 80)
    mask = np.logical_and(seq_len<chunk.shape[1]//10,seq_len>5)
    seq = seq[mask]
    seq_len_filt = seq_len[mask]
    plt.figure()
    plt.hist(seq_len,bins = 80)
    pad_w = np.max(seq_len)
    seq_filt = [map_base(s) for s in seq]
    seq_filt = [np.pad(x,(0,pad_w-len(x))) for x in seq_filt]
    seq_filt = np.asarray(seq_filt)
    print("Write output.")
    os.makedirs(args.output,exist_ok = True)
    chunk_out_f = os.path.join(args.output,'chunks.npy')
    seq_out_f = os.path.join(args.output,'references.npy')
    seq_len_out_f = os.path.join(args.output,'reference_lengths.npy')
    np.save(seq_out_f,seq_filt)
    np.save(seq_len_out_f,seq_len_filt)
    np.save(chunk_out_f,chunk[mask])
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Transfer dataset into bonito-readable format.')
    parser.add_argument('-i', '--input', required = True,
                        help = "The folder contains the chunks,seqs and seq_lens file.")
    parser.add_argument('-o', '--output', required = True,
                        help = "The output folder.")
    args = parser.parse_args(sys.argv[1:])
    main(args)


"""
Created on Tue Aug  3 23:24:39 2021

@author: Haotian Teng
"""
import numpy as np
import os
import sys
import pandas as pd
import argparse
import itertools
from typing import Tuple,Iterable
def load_data(prefix:str,renorm = False,read_meta = False)->Tuple[np.array,Iterable,np.array,np.array,np.array]:
    """
    Load data from the prefix folder

    Parameters
    ----------
    prefix : str
        The folder contains the chunk data and meta information.
    renorm : bool
        If the renomalization chunks is loaded or the orignal chunks.

    Returns
    -------
    chunk : np.array[N,L]
        Contains the N chunks which length L.
    seq : Iterable[N]
        Contains N sequences corresponding.
    seq_lens : TYPE
        DESCRIPTION.
    mms : TYPE
        DESCRIPTION.
    mms_full : TYPE
        DESCRIPTION.

    """
    if renorm:
        chunk_f = os.path.join(prefix,"chunks_renorm.npy")
    else:
        chunk_f = os.path.join(prefix,"chunks.npy")
    seq_f = os.path.join(prefix,"seqs.npy")
    seq_lens_f = os.path.join(prefix,"seq_lens.npy")
    mm_f = os.path.join(prefix,"mm.npy")
    meta_f = os.path.join(prefix,"meta.csv")
    chunk = np.load(chunk_f,mmap_mode = "r")
    seq = np.load(seq_f,mmap_mode = "r")
    seq_lens = np.load(seq_lens_f,mmap_mode = "r")
    if os.path.exists(mm_f) and read_meta:
        mm = np.load(mm_f,allow_pickle=True)
        read_ids = mm[-1]
        metas = pd.read_csv(meta_f,delimiter = " ",header = None)
        metas = metas.dropna(how = 'all')
        ids = list(metas[1])
        mask = [True if id in ids else False for id in read_ids]
        mad,med,meth,offsets,scales,read_ids = [list(itertools.compress(x,mask)) for x in mm]
        mms = np.asarray([mad,med])
        chunk_count = [len(list(j)) for i,j in itertools.groupby(ids)]
        mms_full = np.repeat(mms,chunk_count,axis = 1)
    else:
        mms = None
        mms_full = None
    return chunk,seq,seq_lens,mms,mms_full

def main(args):
    chunk_m,seq_m,seq_lens_m= [],[],[]
    print("Read control dataset.")
    if args.control:
        chunk_c,seq_c,seq_lens_c,mms_c,mms_full_c= load_data(args.control,args.renorm)
    else:
        chunk_c,seq_c,seq_lens_c,mms_c,mms_full_c = [],[],[],[],[]
    print("Read methylation dataset.")
    m_size = 0
    if args.meth:
        for i,m in enumerate(args.meth):
            chunk,seq,seq_lens,mms,mms_full = load_data(m,args.renorm,read_meta = args.shift)
            if args.base_replace:
                print("A -> M transfer.")
                seq = np.array([x.replace('A','M') for x in seq])
            if args.shift:
                offset = np.mean(mms_c[1,:])-np.mean(mms[1,:])
                offset /= mms_full[0,:]
                chunk += offset[:,None]
            chunk_m.append(chunk)
            seq_m.append(seq)
            seq_lens_m.append(seq_lens)
            m_size += chunk.shape[0]
            print(" Methylation dataset %d with shape:%d"%(i+1,chunk.shape[0]))
        print("Control dataset size:%d"%(chunk_c.shape[0]))
        print("Methylation datasets total size:%d"%(m_size))
        if len(args.meth) == 1:
            chunk_m = chunk_m[0]
            seq_m = seq_m[0]
            seq_lens_m = seq_lens_m[0]
        else:
            chunk_m = np.concatenate(chunk_m,axis = 0)
            seq_m = np.concatenate(seq_m,axis = 0)
            seq_lens_m = np.concatenate(seq_lens_m,axis = 0)
    else:
        chunk_m,seq_m,seq_lens_m = [],[],[]
    print("Resize the dataset according to cm_ratio.")
    if args.cm_ratio:
        size_c = len(chunk_c)
        size_m = len(chunk_m)
        curr_ratio = size_c/float(size_m)
        if curr_ratio > args.cm_ratio:
            shrink_size = int(size_m*args.cm_ratio)
            chunk_c,seq_c,seq_lens_c = chunk_c[:shrink_size],seq_c[:shrink_size],seq_lens_c[:shrink_size]
        elif curr_ratio < args.cm_ratio:
            shrink_size = int(size_c/args.cm_ratio)
            chunk_m,seq_m,seq_lens_m = chunk_m[:shrink_size],seq_m[:shrink_size],seq_lens_m[:shrink_size]
    print("Merge control and methylation datasets.")
    if args.control and args.meth:
        chunk_all = np.concatenate([chunk_m,chunk_c],axis = 0)
        seq_all = np.concatenate([seq_m,seq_c],axis = 0)
        seq_lens_all = np.concatenate([seq_lens_m,seq_lens_c],axis = 0)
    elif args.control:
        chunk_all,seq_all,seq_lens_all = chunk_c,seq_c,seq_lens_c
    else:
        chunk_all,seq_all,seq_lens_all = chunk_m,seq_m,seq_lens_m
    out_f = args.output
    print("Save the merged dataset.")
    if args.cm_ratio:
        print("Final size: control - %d, methylation - %d, target size ratio %.1f, final size ratio %.1f"%(len(chunk_c),len(chunk_m),args.cm_ratio,len(chunk_c)/float(len(chunk_m))))
    else:
        print("Final size: control - %d, methylation - %d"%(len(chunk_c),len(chunk_m)))
    os.makedirs(out_f,exist_ok = True)
    np.save(os.path.join(out_f,'chunks.npy'),chunk_all)
    np.save(os.path.join(out_f,'seqs.npy'),seq_all)
    np.save(os.path.join(out_f,'seq_lens.npy'),seq_lens_all)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Training model with tfrecord file')
    parser.add_argument('-m', '--meth', default = None,
                        help = "The methylation folders, multiple directories separate by comma.")
    parser.add_argument('-c', '--control', default = None,
                        help = "The control folder.")
    parser.add_argument('-o', '--output', required = True,
                        help = "The output folder of the merged dataset.")
    parser.add_argument('--shift',action = "store_true", dest = "shift",
                        help = "If move the methylation signal according to the shift of the control median value.")
    parser.add_argument('--cm_ratio',default = None, type = float,
                        help = "The size ratio of control/methylation, size will be adjusted to a maximum reads.")
    parser.add_argument('--no_base_replace', action = "store_false",dest = "base_replace",
                        help = "If the bases in methylated dataset being transferred from A to M.")
    parser.add_argument('--renorm',action="store_true",dest="renorm",
                        help = "Read the renormalization data instead of original data.")
    args = parser.parse_args(sys.argv[1:])
    if not args.control and not args.meth:
        raise ValueError("Neither --control or --meth being specified.")
    if args.cm_ratio and (not args.control or not args.meth):
        raise ValueError("Require both control and methylation dataset being provided when cm_ratio is set.")
    args.meth = args.meth.split(',')
    print(args.base_replace)
    main(args)



#%%
import os
import numpy as np
import seaborn as sns
from xron.xron_train_supervised import main
import tkinter as tk
from tkinter import filedialog

root = tk.Tk()
root.withdraw()

#%% Load data
DF = filedialog.askdirectory()
# DF="/home/heavens/bridge_scratch/ELIGOS_dataset/IVT/control/kmers_guppy_4000_noise/"
chunks = np.load(os.path.join(DF,"chunks.npy"),mmap_mode = "r")
durations = np.load(os.path.join(DF,"durations.npy"),mmap_mode = "r")
kmers = np.load(os.path.join(DF,"kmers.npy"),mmap_mode = "r")
seqs = np.load(os.path.join(DF,"seqs.npy"),mmap_mode = "r")
seq_lens = np.load(os.path.join(DF,"seq_lens.npy"),mmap_mode = "r")

# %% Sanity check
sns.distplot(seq_lens)


"""
Created on Wed Jan 26 12:55:10 2022

@author: Haotian Teng
"""
import os,sys,argparse
from tqdm import tqdm
import h5py
import numpy as np
from xron.utils.seq_op import fast5_iter
from pathlib import Path


def retrive_fastq(seq_h):
    try:
        seq = np.asarray(seq_h['BaseCalled_template']['Fastq']).tobytes().decode('utf-8')
    except:
        seq = str(np.asarray(seq_h['BaseCalled_template']['Fastq']).astype(str))
    return seq

def extract(args):
    iterator = fast5_iter(args.input_fast5,mode = 'r')
    for read_h,signal,fast5_f,read_id in tqdm(iterator):
        seq = retrive_fastq(read_h['Analyses/Basecall_1D_%s'%(args.basecall_entry)])
        with open(os.path.join(args.output,Path(fast5_f).stem+'.fastq'),'a+') as f:
            f.write(seq)
        

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='xron',   
                                     description='A Unsupervised Nanopore basecaller.')
    parser.add_argument('-i', 
                        '--input_fast5', 
                        required = True,
                        help="File path or Folder path to the fast5 file.")
    parser.add_argument('-o',
                        '--output',
                        required = True,
                        help="Output folder.")
    parser.add_argument('--basecall_entry',
                        default = "000",
                        help="The entry number in /Analysis/ to look into, for\
                            example 000 means looking for Basecall_1D_000.")
                            
    FLAGS = parser.parse_args(sys.argv[1:])
    os.makedirs(FLAGS.output,exist_ok = True)
    extract(FLAGS)